unit UBanlistNL;

interface

uses
  Classes, ASGSQLite3, IniFile, Windows, IdHttp, SysUtils, Contnrs,
  IdHttpHeaderInfo, IdTCPConnection, IdComponent, UThreadSafeSQLiteDB,
  IdURI, IdCompressorZLibEX, ZLibExGz, IdSSL;

const
  banlistnl = 'http://banlist.nl/';

type
  TGetProxySettingsEvent = procedure(Sender: TObject; ProxyParams: TIdProxyConnectionInfo) of object;
  TErrorNotifyEvent = procedure (Sender: TObject; Error: String) of object;

  TBanlistNLThread = class(TThread)
  private
    FCancelled: Boolean;
    _Error: String;
    _Progress: Integer;
    DataBase: TWC3BanlistDatabase;
    OnWork: TWorkEvent;
    procedure _DoError;
    procedure _DoProgress;
  public
    CompressedUploads: Boolean;
    OnGetProxySettings: TGetProxySettingsEvent;
    OnProgress: TWorkEvent;
    OnError: TErrorNotifyEvent;
    property Cancelled: Boolean read FCancelled;
    constructor Create(suspended: Boolean);
    procedure Cancel;
  protected
    HTTP: TIdHTTP;
    function GetHTTP(URL: String): String;
    function PostHTTP(URL: String; PostParameters: TStringStream): String;
  end;

  {
    reads the list of online groups from the local database
    and updates the [onlinelistowners] and [onlinegroup_member] tables
    with data from banlist.nl
  }
  TBanlistNLPopulateGroups = class(TBanlistNLThread)
  private
    procedure Work(ASender: TObject; AWorkMode: TWorkMode;
      AWorkCount: Integer);
    procedure ReportError(Job: String; Error: String);
  public
    constructor Create(DataBase: TWC3BanlistDatabase);
    procedure Execute; override;
  end;

  {
    updates the [onlinelistowners] of group 0 (single subscriptions)
  }
  TBanlistNLPopulateMembers = class(TBanlistNLThread)
  private
    procedure Work(ASender: TObject; AWorkMode: TWorkMode;
      AWorkCount: Integer);
    procedure ReportError(Job: String; Error: String);
  public
    constructor Create(DataBase: TWC3BanlistDatabase);
    procedure Execute; override;
  end;

  TBanlistNLDownload = class(TBanlistNLThread)
  private
    procedure Work(ASender: TObject; AWorkMode: TWorkMode;
      AWorkCount: Integer);
    procedure ReportError(Job: String; Error: String);
  public
    User: TBanlistNLMember;
    constructor Create(DataBase: TWC3BanlistDatabase; User: TBanlistNLMember);
    destructor Destroy; override;
    procedure Execute; override;
  end;

  TBanlistNLUpload = class(TBanlistNLThread)
  private
    procedure Work(ASender: TObject; AWorkMode: TWorkMode;
      AWorkCount: Integer);
    procedure ReportError(Job: String; Error: String);
  public
    constructor Create(DataBase: TWC3BanlistDatabase);
    procedure Execute; override;
  end;

implementation

uses DateUtils, Math;

{ TBanlistNLThread }

procedure TBanlistNLThread.Cancel;
begin
  FCancelled := True;
end;

constructor TBanlistNLThread.Create(suspended: Boolean);
begin
  inherited Create(suspended);
  FCancelled := False;
  FreeOnTerminate := True;
  CompressedUploads := False;
end;

function TBanlistNLThread.GetHTTP(URL: String): String;
var
  compressor: TIdCompressorZLibEx;
begin
  http := TIdHTTP.Create(nil);
  compressor := TIdCompressorZLibEx.Create(nil);
  try
    if Assigned(OnGetProxySettings) then
      OnGetProxySettings(self, http.ProxyParams);
    http.OnWork := OnWork;
    http.Compressor := compressor;
    result := http.Get(URL);
  finally
    http.Free;
    compressor.Free;
  end;
end;

function TBanlistNLThread.PostHTTP(URL: String;
  PostParameters: TStringStream): String;
var
  compressor: TIdCompressorZLibEx;
  compressed: TStringStream;
begin
  http := TIdHTTP.Create(nil);
  compressor := TIdCompressorZLibEx.Create(nil);
  try
    if Assigned(OnGetProxySettings) then
      OnGetProxySettings(self, http.ProxyParams);
    http.OnWork := OnWork;
    http.Compressor := compressor;
    http.Request.ContentType := 'application/x-www-form-urlencoded';
    if CompressedUploads then
    begin
      http.Request.ContentEncoding := 'gzip';
      compressed := TStringStream.Create(GZCompressStr(PostParameters.DataString));
      try
        result := http.Post(URL, compressed);
      finally
        compressed.Free;
      end;
    end else
      result := http.Post(URL, PostParameters);
  finally
    http.Free;
    compressor.Free;
  end;
end;

procedure TBanlistNLThread._DoError;
begin
  if Assigned(OnError) then
    OnError(self, _Error);
end;

procedure TBanlistNLThread._DoProgress;
begin
  if Assigned(OnProgress) then
    OnProgress(self, wmRead, _Progress);
end;

{ TBanlistNLPopulateGroups }

constructor TBanlistNLPopulateGroups.Create(
  DataBase: TWC3BanlistDatabase);
begin
  inherited Create(true);
  self.DataBase := DataBase;
  self.OnWork := Work;
end;

procedure TBanlistNLPopulateGroups.Execute;
var
  groups: TObjectList;
  i: Integer;
  g: TBanlistNLGroup;
  sl: TStringList;
begin
  groups := TObjectList.Create;
  sl := TStringList.Create;
  try
    DataBase.EnumerateBanlistNLGroups(groups);
  except
    on e: Exception do ReportError('enumerating banlist.nl groups', e.Message);
  end;
  try
    for i := 0 to groups.Count - 1 do
    begin
      _Progress := Round(i / groups.Count * 100);
      if Assigned(OnProgress) then Synchronize(_DoProgress);
      g := TBanlistNLGroup(groups[i]);
      try
        sl.Text := GetHTTP(banlistnl + 'bl_server.php?g=' + IntToStr(g.OnlineID));
        try
          DataBase.ParseGroupMembers(sl, g.ID);
        except
          on e: Exception do ReportError('parsing server reply for group ' + g.Name, e.Message);
        end;
      except
        on e: Exception do ReportError('getting information on group ' + g.Name, e.Message);
      end;
    end;
  finally
    groups.Free;
    sl.Free;
    _Progress := 100;
    if Assigned(OnProgress) then Synchronize(_DoProgress);
  end;
end;

procedure TBanlistNLPopulateGroups.ReportError(Job, Error: String);
begin
  if Assigned(OnError) then
  begin
    _Error := 'Error while ' + Job + ':' + Error;
    Synchronize(_DoError);
  end;
end;

procedure TBanlistNLPopulateGroups.Work(ASender: TObject;
  AWorkMode: TWorkMode; AWorkCount: Integer);
begin
  if Cancelled then
    HTTP.Disconnect;
end;

{ TBanlistNLPopulateMembers }

constructor TBanlistNLPopulateMembers.Create(
  DataBase: TWC3BanlistDatabase);
begin
  inherited Create(true);
  self.DataBase := DataBase;
  self.OnWork := Work;
end;

procedure TBanlistNLPopulateMembers.Execute;
var
  members: TObjectList;
  sl: TStringList;
  i: Integer;
  m: TBanlistNLMember;
  a: String;
begin
  members := TObjectList.Create;
  sl := TStringList.Create;
  try
    DataBase.EnumerateBanlistNLMembers(members, 0);
  except
    on e: Exception do ReportError('enumerating banlist.nl members', e.Message);
  end;

  try
    a := banlistnl + 'bl_server.php?';
    if members.Count > 0 then
    begin
      for i := 0 to members.Count - 1 do
      begin
        m := TBanlistNLMember(members[i]);
        a := a + 'u[]=' + TIdURI.ParamsEncode(m.Name) + '&';
      end;
      try
        sl.Text := GetHTTP(a);
        try
          DataBase.ParseGroupMembers(sl, 0);
        except
          on e: Exception do ReportError('parsing server reply for subscribed banlist.nl users', e.Message);
        end;
      except
        on e: Exception do ReportError('getting information on subscribed banlist.nl users', e.Message);
      end;
    end;
  finally
    members.Free;
    sl.Free;
    _Progress := 100;
    if Assigned(OnProgress) then Synchronize(_DoProgress);
  end;
end;

procedure TBanlistNLPopulateMembers.ReportError(Job, Error: String);
begin
  if Assigned(OnError) then
  begin
    _Error := 'Error while ' + Job + ':' + Error;
    Synchronize(_DoError);
  end;
end;

procedure TBanlistNLPopulateMembers.Work(ASender: TObject;
  AWorkMode: TWorkMode; AWorkCount: Integer);
begin
  if Cancelled then
    HTTP.Disconnect;
end;

{ TBanlistNLDownload }

constructor TBanlistNLDownload.Create(DataBase: TWC3BanlistDatabase;
  User: TBanlistNLMember);
begin
  inherited Create(true);
  self.DataBase := DataBase;
  self.OnWork := Work;
  self.User := TBanlistNLMember.Create;
  self.User.Assign(User);
end;

destructor TBanlistNLDownload.Destroy;
begin
  User.Free;
  inherited;
end;

procedure TBanlistNLDownload.Execute;
var
  sl: TStringList;
begin
  sl := TStringList.Create;
  try
    try
      sl.Text := GetHTTP(User.URL + '?u=' + IntToStr(User.OnlineID));
      try
        DataBase.ParseBansOfMember(sl, User);
      except
        on e: Exception do ReportError('parsing server reply for the download of user', e.Message);
      end;
    except
      on e: Exception do ReportError('downloading the list of user', e.Message);
    end;
  finally
    sl.Free;
  end;
  _Progress := 100;
  if Assigned(OnProgress) then Synchronize(self, _DoProgress);
end;

procedure TBanlistNLDownload.ReportError(Job, Error: String);
begin
  if Assigned(OnError) then
  begin
    _Error := 'Error while ' + Job + ':' + Error;
    Synchronize(_DoError);
  end;
end;

procedure TBanlistNLDownload.Work(ASender: TObject; AWorkMode: TWorkMode;
  AWorkCount: Integer);
begin
  if Cancelled then
    HTTP.Disconnect;
end;

{ TBanlistNLUpload }

constructor TBanlistNLUpload.Create(DataBase: TWC3BanlistDatabase);
begin
  inherited Create(true);
  self.DataBase := DataBase;
  self.OnWork := Work;
  self.FreeOnTerminate := True;
end;

procedure TBanlistNLUpload.Execute;
var
  sl: TStringList;
  LocalList: TObjectList;
  PlayerList: TObjectList;
  i, j: Integer;
  l: TLocalEntry;
  p: TPlayerEntry;
  Params: TStringStream;
  r: String;
begin
  sl := TStringList.Create;
  LocalList := TObjectList.Create;
  PlayerList := TObjectList.Create;
  Params := TStringStream.Create('');
  try
    try
      Database.EnumerateBans(LocalList);
      Database.EnumeratePlayers(PlayerList);
      for i := 0 to LocalList.Count - 1 do
      begin
        l := TLocalEntry(LocalList[i]);
        for j := 0 to PlayerList.Count - 1 do
        begin
          p := TPlayerEntry(PlayerList[i]);
          if l.PlayerID = p.ID then
          begin
            sl.Add(p.Name + #9 + IntToStr(p.Realm shl 4 + l.Typ) + #9 + IntToStr(Ord(l.Privat)) + #9 + l.Comment);
            break;
          end;
        end;
      end;
      Params.WriteString('username=' + TIdURI.ParamsEncode(Database.GetSetting('banlistnl.username')) + '&');
      Params.WriteString('password=' + TIdURI.ParamsEncode(Database.GetSetting('banlistnl.password')) + '&');
      Params.WriteString('bans=' + TIdURI.ParamsEncode(sl.Text));
      try
        r := PostHTTP(banlistnl + 'bl_upload.php', Params);
        if r <> '' then
          ReportError('banlist.nl was processing uploaded the local database', r); 
      except
        on e: Exception do ReportError('uploading the local database', e.Message);
      end;
    except
      on e: Exception do ReportError('building data for upload of the local database', e.Message);
    end;
  finally
    sl.Free;
    LocalList.Free;
    PlayerList.Free;
    Params.Free;
  end;
  _Progress := 100;
  if Assigned(OnProgress) then Synchronize(self, _DoProgress);
end;

procedure TBanlistNLUpload.ReportError(Job, Error: String);
begin
  if Assigned(OnError) then
  begin
    _Error := 'Error while ' + Job + ':' + #13#10 + Error;
    Synchronize(_DoError);
  end;
end;

procedure TBanlistNLUpload.Work(ASender: TObject; AWorkMode: TWorkMode;
  AWorkCount: Integer);
begin
  if Cancelled then
    HTTP.Disconnect;
end;

end.

