unit UThreadSafeSQLiteDB;

interface

  uses Windows, ASGSQLite3, Classes, Contnrs, Inifile, SysUtils, DateUtils,
    Variants;

  type
    TBanlistNLEntry = class(TObject)
    public
      Name: String;
      Typ: Byte;
      Comment: String;
    end;

    TBanlistNLGroup = class(TObject)
    public
      ID: Integer;
      OnlineID: Integer;
      Active: Boolean;
      Name: String;
      Icon: String;
    end;

    TBanlistNLMember = class(TPersistent)
    public
      ID: Integer;
      OnlineID: Integer;
      Active: Boolean;
      Name: String;
      DateTimeLocal: TDateTime;
      DateTimeRemote: TDateTime;
      CountRemote: Integer;
      URL: String;
      procedure Assign(Source: TPersistent); override;
    end;

    TLocalEntry = class(TPersistent)
    public
      ID: Integer;
      PlayerID: Integer;
      Comment: String;
      LastChange: TDateTime;
      Privat: Boolean;
      Typ: Byte;
      procedure Assign(Source: TPersistent); override;
    end;

    TPlayerEntry = class(TPersistent)
    public
      ID: Integer;
      Name: String;
      LastChange: TDateTime;
      Realm: Integer;
    end;

    TASQLite3DBThreadsafe = class(TASQLite3DB)
    public
      Mutex: THandle;
      procedure Lock;
      procedure Unlock;
      function IsNull(v: Variant; valueIfNull: Variant): Variant;
      constructor Create(AOwner: TComponent); override;
    end;

    TWC3BanlistDatabase = class(TASQLite3DBThreadsafe)
    public
      procedure EnumerateBanlistNLGroups(l: TObjectList);
      procedure EnumerateBanlistNLMembers(l: TObjectList; groupID: Integer = -1);
      procedure EnumerateBans(l: TObjectList);
      procedure EnumeratePlayers(l: TObjectList);
      procedure ParseGroupMembers(sl: TStringList; groupID: Integer);
      procedure ParseBansOfMember(sl: TStringList; User: TBanlistNLMember);
      function GetSetting(Name: String): String;
      procedure SetSetting(Name: String; Value: String);
    end;

  function ISOTimeToDateTime(ISOTime: String): TDateTime;

implementation

uses DB;

function ISOTimeToDateTime(ISOTime: String): TDateTime;
var
  y, mm, d: word;
  h, m: word;
begin
  // 20080403T1533Z
  y := StrToInt(Copy(ISOTime, 1, 4));
  mm := StrToInt(Copy(ISOTime, 5, 2));
  d := StrToInt(Copy(ISOTime, 7, 2));
  h := StrToInt(Copy(ISOTime, 10, 2));
  m := StrToInt(Copy(ISOTime, 12, 2));
  result := EncodeDateTime(y, mm, d, h, m, 0, 0);
end;

{ TASQLite3DBThreadsafe }

constructor TASQLite3DBThreadsafe.Create;
begin
  inherited Create(AOwner);
  Mutex := CreateMutex(nil, false, nil);
end;

function TASQLite3DBThreadsafe.IsNull(v, valueIfNull: Variant): Variant;
begin
  if VarIsNull(v) then
    result := valueIfNull
  else
    result := v;
end;

procedure TASQLite3DBThreadsafe.Lock;
begin
  WaitForSingleObject(Mutex, INFINITE);
end;

procedure TASQLite3DBThreadsafe.Unlock;
begin
  ReleaseMutex(Mutex);
end;

{ TWC3BanlistDatabase }

procedure TWC3BanlistDatabase.EnumerateBanlistNLGroups(l: TObjectList);
var
  g: TBanlistNLGroup;
  sql: TASQLite3Query;
begin
  sql := TASQLite3Query.Create(self);
  Lock;
  try
    sql.Connection := self;
    sql.SQL.Text := 'SELECT id, onlineid, active, name, icon FROM onlinegroups';
    sql.ExecSQL;
    sql.First;
    while not sql.Eof do
    begin
      g := TBanlistNLGroup.Create;
      g.ID := sql['id'];
      g.OnlineID := sql['onlineid'];
      g.Active := sql['active'] = 1;
      g.Name := sql['name'];
      g.Icon := sql['icon'];
      l.Add(g);
      sql.Next;
    end;
  finally
    Unlock;
    sql.Free;
  end;
end;

procedure TWC3BanlistDatabase.EnumerateBanlistNLMembers(l: TObjectList; groupID: Integer = -1);
var
  m: TBanlistNLMember;
  sql: TASQLite3Query;
begin
  sql := TASQLite3Query.Create(self);
  Lock;
  try
    sql.Connection := self;
    sql.SQL.Text := 'SELECT OLO.ID, OLO.onlineID, OLO.owner, OLO.Active, OLO.DateTimeLocal, ' +
      'OLO.DateTimeRemote, OLO.CountRemote, OLO.URL FROM onlinelistowners AS OLO';
    if groupID <> -1 then
    begin
      sql.SQL.Text := sql.SQL.Text + ' INNER JOIN onlinegroup_member AS OGM ON ' +
        'OLO.id = OGM.ownerID WHERE ogm.groupID = :groupID';
      sql.Params.ParamByName('groupID').Value := groupID;
    end;
    sql.ExecSQL;
    sql.First;
    while not sql.Eof do
    begin
      m := TBanlistNLMember.Create;
      m.ID := sql['id'];
      m.OnlineID := IsNull(sql['onlineID'], 0);
      m.Name := sql['owner'];
      m.Active := IsNull(sql['active'], 0) = 1;
      m.DateTimeLocal := IsNull(sql['datetimelocal'], 0);
      m.DateTimeRemote := IsNull(sql['datetimeremote'], 0);
      m.CountRemote := IsNull(sql['countremote'], 0);
      m.URL := IsNull(sql['url'], '');
      l.Add(m);
      sql.Next;
    end;
  finally
    Unlock;
    sql.Free;
  end;
end;

procedure TWC3BanlistDatabase.EnumerateBans(l: TObjectList);
var
  sql: TASQLite3Query;
  e: TLocalEntry;
begin
  sql := TASQLite3Query.Create(self);
  try
    Lock;
    sql.Connection := self;
    sql.SQL.Text := 'SELECT * FROM bans';
    sql.ExecSQL;
    sql.First;
    while not sql.Eof do
    begin
      e := TLocalEntry.Create;
      e.ID := sql.FieldValues['id'];
      e.PlayerID := sql.FieldValues['playerid'];
      e.Comment := sql.FieldValues['comment'];
      e.LastChange := sql.FieldValues['datetime'];
      e.Privat := sql.FieldValues['private'] = 1;
      e.Typ := sql.FieldValues['type'];
      l.Add(e);
      sql.Next;
    end;
  finally
    Unlock;
    sql.Free;
  end;
end;

procedure TWC3BanlistDatabase.EnumeratePlayers(l: TObjectList);
var
  sql: TASQLite3Query;
  e: TPlayerEntry;
begin
  sql := TASQLite3Query.Create(self);
  try
    Lock;
    sql.Connection := self;
    sql.SQL.Text := 'SELECT * FROM players';
    sql.ExecSQL;
    sql.First;
    while not sql.Eof do
    begin
      e := TPlayerEntry.Create;
      e.ID := sql.FieldValues['id'];
      e.Name := sql.FieldValues['nick'];
      e.LastChange := sql.FieldValues['datetime'];
      e.Realm := sql.FieldValues['realm'];
      l.Add(e);
      sql.Next;
    end;
  finally
    Unlock;
    sql.Free;
  end;
end;

function TWC3BanlistDatabase.GetSetting(Name: String): String;
var
  sql: TASQLite3Query;
begin
  result := '';
  sql := TASQLite3Query.Create(self);
  try
    Lock;
    sql.Connection := self;
    sql.SQL.Text := 'SELECT value FROM settings WHERE name = :setting';
    sql.Params.ParamByName('setting').Value := Name;
    sql.ExecSQL;
    result := sql.FieldValues['value'];
  finally
    Unlock;
    sql.Free;
  end;
end;

procedure TWC3BanlistDatabase.ParseBansOfMember(sl: TStringList;
  User: TBanlistNLMember);
var
  row: TStringList;
  sql: TASQLite3Query;
  i: Integer;
begin
  row := TStringList.Create;
  sql := TASQLite3Query.Create(self);
  try
    Lock;
    sql.Connection := self;
    sql.SQL.Text := 'DELETE FROM onlinelist WHERE ownerid = :ownerid';
    sql.Params.ParamByName('ownerid').Value := User.ID;
    sql.ExecSQL;
    for i := 0 to sl.Count - 1 do
    begin
      if sl[i] <> '' then
      begin
        Split(sl[i], #9, row);
        sql.SQL.Text := 'INSERT INTO onlinelist (ownerid, nickname, comment, type) VALUES (:ownerid, :nickname, :comment, :type)';
        sql.Params.ParamByName('ownerid').Value := user.ID;
        sql.Params.ParamByName('nickname').Value := row[0];
        sql.Params.ParamByName('comment').Value := row[1];
        sql.Params.ParamByName('type').Value := row[2];
        sql.ExecSQL;
      end;
    end;
    sql.SQL.Text := 'UPDATE onlinelistowners SET datetimeLocal = :datetimeLocal WHERE id = :id';
    sql.Params.ParamByName('datetimeLocal').Value := Now;
    user.DateTimeLocal := now;
    sql.Params.ParamByName('id').Value := user.ID;
    sql.ExecSQL;
  finally
    Unlock;
    row.Free;
    sql.Free;
  end;
end;

procedure TWC3BanlistDatabase.ParseGroupMembers(sl: TStringList; groupID: Integer);
var
  sql: TASQLite3Query;
  row: TStringList;
  i: Integer;

  name: String;
  onlineID: Integer;
  datetimeRemote: TDateTime;
  countRemote: Integer;
  url: String;

  ids: Array of Integer;
begin
  setlength(ids, 0);
  sql := TASQLite3Query.Create(nil);
  row := TStringList.Create;
  try
    Lock;
    sql.Connection := self;
    for i := 0 to sl.Count - 1 do
    begin
      Split(sl[i], #9, row);
      if row[2] <> '' then
      begin
        name := row[0];
        onlineID := StrToInt(row[1]);
        datetimeRemote := ISOTimeToDateTime(row[2]);
        countRemote := StrToInt(row[3]);
        url := row[4];
        setlength(ids, length(ids) + 1);

        sql.SQL.Text := 'SELECT id, onlineid, owner, active, datetimeremote, countremote, url FROM onlinelistowners WHERE onlineid = :onlineid';
        sql.Params.ParamByName('onlineid').Value := onlineid;
        sql.ExecSQL;
        if sql.Eof then
        begin
          sql.SQL.Text := 'INSERT INTO onlinelistowners (onlineid, owner, active, datetimeremote, countremote, url) VALUES ' +
            '(:onlineid, :owner, :active, :datetimeremote, :countremote, :url)';
          sql.Params.ParamByName('active').Value := 1;
          sql.Params.ParamByName('owner').Value := name;
          sql.Params.ParamByName('onlineid').Value := onlineid;
          sql.Params.ParamByName('datetimeremote').Value := datetimeremote;
          sql.Params.ParamByName('countremote').Value := countremote;
          sql.Params.ParamByName('url').Value := url;
          sql.ExecSQL;
        end else
        begin
          sql.SQL.Text := 'UPDATE onlinelistowners SET owner = :owner, onlineid = :onlineid, datetimeremote = :datetimeremote, countremote = :countremote, url = :url WHERE onlineid = :onlineid';
          sql.Params.ParamByName('owner').Value := name;
          sql.Params.ParamByName('onlineid').Value := onlineid;
          sql.Params.ParamByName('datetimeremote').Value := datetimeremote;
          sql.Params.ParamByName('countremote').Value := countremote;
          sql.Params.ParamByName('url').Value := url;
          sql.ExecSQL;
        end;

        sql.SQL.Text := 'SELECT id, onlineid, owner, active, datetimeremote, countremote, url FROM onlinelistowners WHERE onlineid = :onlineid';
        sql.Params.ParamByName('onlineid').Value := onlineid;
        sql.ExecSQL;
        sql.First;
        ids[length(ids)-1] := sql.FieldValues['id'];
      end;
    end;
    sql.SQL.Text := 'DELETE FROM onlinegroup_member WHERE groupid = :groupid';
    sql.Params.ParamByName('groupid').Value := groupid;
    sql.ExecSQL;
    for i := 0 to length(ids)-1 do
    begin
      sql.SQL.Text := 'INSERT INTO onlinegroup_member (groupid, ownerid) VALUES (:groupid, :ownerid)';
      sql.Params.ParamByName('groupid').Value := groupid;
      sql.Params.ParamByName('ownerid').Value := ids[i];
      sql.ExecSQL;
    end;
  finally
    Unlock;
    row.Free;
    sql.Free;
  end;
end;

procedure TWC3BanlistDatabase.SetSetting(Name, Value: String);
var
  sql: TASQLite3Query;
begin
  sql := TASQLite3Query.Create(self);
  try
    Lock;
    sql.Connection := self;
    sql.SQL.Text := 'SELECT COUNT(*) AS anzahl FROM settings WHERE name = :name';
    sql.Params.ParamByName('name').Value := Name;
    if sql.FieldValues['anzahl'] > 0 then
    begin
      sql.SQL.Text := 'UPDATE settings SET value = :value WHERE name = :name';
      sql.Params.ParamByName('value').Value := Value;
      sql.Params.ParamByName('name').Value := Name;
      sql.ExecSQL;
    end else
    begin
      sql.SQL.Text := 'INSERT INTO settings (name, value) VALUES (:name, :value)';
      sql.Params.ParamByName('value').Value := Value;
      sql.Params.ParamByName('name').Value := Name;
      sql.ExecSQL;
    end;
  finally
    Unlock;
    sql.Free;
  end;
end;

{ TBanlistNLMember }

procedure TBanlistNLMember.Assign(Source: TPersistent);
begin
  if Source is TBanlistNLMember then
  begin
    ID := TBanlistNLMember(Source).ID;
    OnlineID := TBanlistNLMember(Source).OnlineID;
    Active := TBanlistNLMember(Source).Active;
    Name := TBanlistNLMember(Source).Name;
    DateTimeLocal := TBanlistNLMember(Source).DateTimeLocal;
    DateTimeRemote := TBanlistNLMember(Source).DateTimeRemote;
    CountRemote := TBanlistNLMember(Source).CountRemote;
    URL := TBanlistNLMember(Source).URL;
  end else
    inherited Assign(Source);
end;

{ TBanlistNLLocalEntry }

procedure TLocalEntry.Assign(Source: TPersistent);
begin
  if Source is TLocalEntry then
  begin
    ID := TLocalEntry(Source).ID;
    PlayerID := TLocalEntry(Source).PlayerID;
    Comment := TLocalEntry(Source).Comment;
    LastChange := TLocalEntry(Source).LastChange;
    Privat := TLocalEntry(Source).Privat;
    Typ := TLocalEntry(Source).Typ;
  end else
    inherited Assign(Source);
end;

end.
