unit PingPlotDrawer;

interface

uses
  SysUtils, Windows, Classes, Graphics, FiveNum;

type
  TPingPlotMode = (pmNormal, pmSimple);

  TPingPlotPlayer = record
    Name: string;
    Color: TColor;
  end;
  TPingPlotPlayers = array of TPingPlotPlayer;

  TPingPlotValue = record
    Summary: TFiveNumberSummary;
    TopWhisker: Integer;
    BottomWhisker: Integer;
    Mean: Integer;
    ExtremeOutliers: TList;
    MildOutliers: TList;
  end;
  TPingPlotValues = array of TPingPlotValue;

  TPingPlotDrawer = class
  private
    BrightColors, DarkColors: array of TColor;
    Canvas: TCanvas;
    Count: Integer;
    Mode: TPingPlotMode;
    Players: TPingPlotPlayers;
    Values: TPingPlotValues;
    minV, maxV: Integer;
  public
    constructor Create(aCanvas: TCanvas);
    destructor Destroy; override;
    procedure Initialize(aPlayers: TPingPlotPlayers; aMode: TPingPlotMode);
      virtual;
    procedure Data(aPlayer: Integer; aData: TList); virtual;
    procedure Draw(DrawSpace: TRect); virtual;
    procedure DrawBackground(DrawSpace: TRect); virtual;
  end;

implementation

uses
  Math;

const
  EmptyValue: TPingPlotValue = (
    Summary: (
      Minimum: -1;
      LowerQuartile: -1;
      Median: -1;
      UpperQuartile: -1;
      Maximum: -1
    );
    TopWhisker: -1;
    BottomWhisker: -1;
    Mean: -1;
    ExtremeOutliers: nil;
    MildOutliers: nil
  );

function BlendScreen(Color1, Color2: TColor): TColor;
var
  r1,r2,g1,g2,b1,b2: byte;
begin
  r1 := (not (Color1 and $FF));
  g1 := (not ((Color1 and $FF00) shr 8));
  b1 := (not ((Color1 and $FF0000) shr 16));
  r2 := (not (Color2 and $FF));
  g2 := (not ((Color2 and $FF00) shr 8));
  b2 := (not ((Color2 and $FF0000) shr 16));
  r1 := (not ((r1 * r2) div 255));
  g1 := (not ((g1 * g2) div 255));
  b1 := (not ((b1 * b2) div 255));
  Result := TColor(r1 or (g1 shl 8) or (b1 shl 16));
end;

function BlendMultiply(Color1, Color2: TColor): TColor;
var
  r1,r2,g1,g2,b1,b2: byte;
begin
  r1 := (Color1 and $FF);
  g1 := ((Color1 and $FF00) shr 8);
  b1 := ((Color1 and $FF0000) shr 16);
  r2 := (Color2 and $FF);
  g2 := ((Color2 and $FF00) shr 8);
  b2 := ((Color2 and $FF0000) shr 16);
  r1 := ((r1 * r2) div 255);
  g1 := ((g1 * g2) div 255);
  b1 := ((b1 * b2) div 255);
  Result := TColor(r1 or (g1 shl 8) or (b1 shl 16));
end;

{ TPingPlotDrawer }

constructor TPingPlotDrawer.Create(aCanvas: TCanvas);
begin
  inherited Create;

  Canvas := aCanvas;
  Count := 0;
  Mode := pmNormal;
  minV := 1000000000;
  maxV := -1000000000;
end;

procedure TPingPlotDrawer.Data(aPlayer: Integer; aData: TList);
var
  i, iq, tmin, tmax: Integer;
begin
  if (aPlayer < 0) or (aPlayer > Count - 1) then
    raise Exception.Create('PingPlotDrawer: Invalid player number!');

  if (aData = nil) or (aData.Count = 0) then
  begin
    if Values[aPlayer].ExtremeOutliers <> nil then
      Values[aPlayer].ExtremeOutliers.Free;
    if Values[aPlayer].MildOutliers <> nil then
      Values[aPlayer].MildOutliers.Free;
    Values[aPlayer] := EmptyValue;
  end
  else
    with Values[aPlayer] do
    begin
      Summary := FiveNumberSummary(SortList(aData));
      if Mode = pmNormal then
      begin
        Mean := 0;
        for i := 0 to aData.Count - 1 do
          Inc(Mean, Integer(aData.Items[i]));
        Mean := Mean div aData.Count;


        iq := Summary.UpperQuartile - Summary.LowerQuartile;

        if ExtremeOutliers = nil then
          ExtremeOutliers := TList.Create
        else
          ExtremeOutliers.Clear;

        if MildOutliers = nil then
          MildOutliers := TList.Create
        else
          MildOutliers.Clear;

        i := 0;
        while Integer(aData.Items[i]) < Summary.LowerQuartile - iq * 1.5 do
        begin
          if Integer(aData.Items[i]) < Summary.LowerQuartile - iq * 3.0 then
            ExtremeOutliers.Add(aData.Items[i])
          else
            MildOutliers.Add(aData.Items[i]);
          Inc(i);
        end;
        BottomWhisker := Integer(aData.Items[i]);

        i := aData.Count - 1;
        while Integer(aData.Items[i]) > Summary.UpperQuartile + iq * 1.5 do
        begin
          if Integer(aData.Items[i]) > Summary.UpperQuartile + iq * 3.0 then
            ExtremeOutliers.Add(aData.Items[i])
          else
            MildOutliers.Add(aData.Items[i]);
          Dec(i);
        end;
        TopWhisker := Integer(aData.Items[i]);
      end;
    end;

  tmin := 1000000000;
  tmax := -1000000000;

  for i := 0 to Count - 1 do
  begin
    if Values[i].Summary.Maximum = -1 then
      Continue;

    tmin := Min(tmin, (Values[i].Summary.Minimum div 100) * 100);
    tmax := Max(tmax, (Values[i].Summary.Maximum div 100 + 1) * 100);
  end;

  if (tmin < minV) or (tmin > (minV + 199)) then
    minV := tmin;

  if (tmax > maxV) or (tmax < (maxV - 199)) then
    maxV := tmax;

end;

destructor TPingPlotDrawer.Destroy;
var
  i: integer;
begin
  if Count > 0 then
  for i := 0 to Count - 1 do
  begin
    if Values[i].ExtremeOutliers <> nil then
      Values[i].ExtremeOutliers.Free;
    if Values[i].MildOutliers <> nil then
      Values[i].MildOutliers.Free;
  end;

  SetLength(BrightColors, 0);
  SetLength(DarkColors, 0);
  SetLength(Values, 0);

  inherited Destroy;
end;

procedure TPingPlotDrawer.Draw(DrawSpace: TRect);
var
  i, j, SlotHeight, LeftMargin, TextSize, PlotWidth: Integer;
  TickCount: Integer;
  a,b: Integer;
  r1,r2: TRect;
  s: string;
begin
  DrawBackground(DrawSpace);

  if (Count = 0) then
    Exit;

  SlotHeight := (DrawSpace.Bottom - DrawSpace.Top - 64) div Count;

  with Canvas do
  begin
    Font.Name := 'Tahoma';
    Font.Size := 8;
    Font.Style := [fsBold];
    TextSize := TextHeight('X');

    LeftMargin := DrawSpace.Left + 45;

    for i := 0 to Count - 1 do
    begin
      LeftMargin := Max(LeftMargin, DrawSpace.Left + 45 +
        TextWidth(Players[i].Name));

      if ((minV <> 1000000000) or (maxV <> -1000000000))
      and (SlotHeight > TextSize * 2 + 2) then
      begin
        s := 'Med: ' + IntToStr(Values[i].Summary.Median) + 'ms Max: ' +
              IntToStr(Values[i].Summary.Maximum) + 'ms';
        Font.Style := [];
        LeftMargin := Max(LeftMargin, DrawSpace.Left + 45 +
          TextWidth(s));
        Font.Style := [fsBold];
      end;
    end;

    if ((minV <> 1000000000) or (maxV <> -1000000000))
    and (SlotHeight > TextSize * 2 + 2) then
      a := DrawSpace.Top + 32 + SlotHeight shr 1 - TextSize
    else
      a := DrawSpace.Top + 32 + (SlotHeight - TextSize) shr 1;

    r1.Left := DrawSpace.Left;
    r1.Right := DrawSpace.Left + 20;
    r2.Left := DrawSpace.Left + 20;
    r2.Right := DrawSpace.Right;

    for i := 0 to Count - 1 do
    begin
      Brush.Color := Players[i].Color;
      r1.Top := DrawSpace.Top + 32 + SlotHeight * i;
      r1.Bottom := DrawSpace.Top + 32 + SlotHeight * (i + 1);
      FillRect(r1);

      Brush.Color := BrightColors[i];
      r2.Top := DrawSpace.Top + 32 + SlotHeight * i;
      r2.Bottom := DrawSpace.Top + 32 + SlotHeight * (i + 1);
      FillRect(r2);

      Font.Color := DarkColors[i];
      TextOut(DrawSpace.Left + 22, a + SlotHeight * i, Players[i].Name);
    end;

    if (minV <> 1000000000) or (maxV <> -1000000000) then
    begin
      Brush.Style := bsClear;
      Pen.Width := 1;
      Pen.Color := clBlack;
      Pen.Style := psSolid;
      Font.Style := [];
      TextSize := TextHeight('X');

      if SlotHeight > TextSize * 2 + 2 then
        for i := 0 to Count - 1 do
        begin
          Font.Color := DarkColors[i];
          b := DrawSpace.Top + 32 + SlotHeight * i + SlotHeight shr 1;
          s := 'Med: ' + IntToStr(Values[i].Summary.Median) + 'ms Max: ' +
            IntToStr(Values[i].Summary.Maximum) + 'ms';
          TextOut(DrawSpace.Left + 22, b, s);
        end;

      if LeftMargin < 0 then
        Exit;

      Font.Color := clBlack;

      PlotWidth := DrawSpace.Right - LeftMargin - 15;

      TickCount := PlotWidth div 35;

      while (TickCount > 0) and (Frac((maxV - minV) / TickCount / 5) > 0) do
        Dec(TickCount);

      if TickCount <= 0 then
        Exit;

      a := DrawSpace.Top + 32 - 5;
      b := a + SlotHeight * Count + 10;

      MoveTo(LeftMargin, a);
      LineTo(LeftMargin + PlotWidth + 1, a);
      MoveTo(LeftMargin, b);
      LineTo(LeftMargin + PlotWidth + 1, b);
      for i := 0 to TickCount do
      begin
        MoveTo(LeftMargin + PlotWidth * i div TickCount, a - 10);
        LineTo(LeftMargin + PlotWidth * i div TickCount, a);
        MoveTo(LeftMargin + PlotWidth * i div TickCount, b + 10);
        LineTo(LeftMargin + PlotWidth * i div TickCount, b);

        s := IntToStr((maxV - minV) * i div TickCount + minV);
        TextOut(LeftMargin + PlotWidth * i div TickCount - TextWidth(s) shr 1,
          a - 12 - TextSize, s);
        TextOut(LeftMargin + PlotWidth * i div TickCount - TextWidth(s) shr 1,
          b + TextSize, s);
      end;

      for i := 0 to Count - 1 do
      begin
        if Values[i].Summary.Maximum = -1 then
          Continue;

        b := DrawSpace.Top + 32 + SlotHeight * i + SlotHeight shr 1;

        Brush.Style := bsClear;
        Pen.Style := psSolid;
        Pen.Color := DarkColors[i];

        r1.Top := DrawSpace.Top + 32 + SlotHeight * i + 2;
        r1.Left := LeftMargin + ((Values[i].Summary.LowerQuartile - minV) * PlotWidth)
          div (maxV - minV);
        r1.Right := LeftMargin + ((Values[i].Summary.UpperQuartile - minV) * PlotWidth)
          div (maxV - minV) + 1;
        r1.Bottom := DrawSpace.Top + 32 + SlotHeight * (i + 1) - 2;
        Rectangle(r1);

        Pen.Width := 3;
        a := LeftMargin + ((Values[i].Summary.Median - minV) * PlotWidth) div
          (maxV - minV);
        MoveTo(a, r1.Top - 1);
        LineTo(a, r1.Bottom);

        Pen.Width := 1;

        if Mode = pmNormal then
          a := LeftMargin + ((Values[i].BottomWhisker - minV) * PlotWidth) div
            (maxV - minV)
        else
          a := LeftMargin + ((Values[i].Summary.Minimum - minV) * PlotWidth) div
            (maxV - minV);

        MoveTo(a, r1.Top);
        LineTo(a, r1.Bottom);
        Pen.Style := psDot;
        MoveTo(a, b);
        LineTo(r1.Left, b);

        Pen.Style := psSolid;
        if Mode = pmNormal then
          a := LeftMargin + ((Values[i].TopWhisker - minV) * PlotWidth) div
            (maxV - minV)
        else
          a := LeftMargin + ((Values[i].Summary.Maximum - minV) * PlotWidth) div
            (maxV - minV);

        MoveTo(a, r1.Top);
        LineTo(a, r1.Bottom);
        Pen.Style := psDot;
        MoveTo(a, b);
        LineTo(r1.Right, b);

        if Mode = pmNormal then
        begin
          Pen.Style := psSolid;
          Brush.Color := DarkColors[i];
          Brush.Style := bsSolid;

          a := LeftMargin + ((Values[i].Mean - minV) * PlotWidth) div
            (maxV - minV);
          Ellipse(a - 1, b - 1, a + 2, b + 2);

          with Values[i].MildOutliers do
            for j := 0 to Count - 1 do
            begin
              a := LeftMargin + ((Integer(Items[j]) - minV) * PlotWidth) div
                (maxV - minV);
              Ellipse(a - 2, b - 2, a + 3, b + 3);
            end;

          Brush.Style := bsClear;

          with Values[i].ExtremeOutliers do
            for j := 0 to Count - 1 do
            begin
              a := LeftMargin + ((Integer(Items[j]) - minV) * PlotWidth) div
                (maxV - minV);
              Ellipse(a - 2, b - 2, a + 3, b + 3);
            end;
        end;
      end;
    end;
  end;
end;

procedure TPingPlotDrawer.DrawBackground(DrawSpace: TRect);
begin
  with Canvas do
  begin
    Pen.Width := 0;
    Brush.Style := bsSolid;
    Brush.Color := clWhite;
    Canvas.FillRect(DrawSpace);
  end;
end;

procedure TPingPlotDrawer.Initialize(aPlayers: TPingPlotPlayers;
  aMode: TPingPlotMode);
var
  i: Integer;
begin
  if Count > 0 then
    for i := 0 to Count - 1 do
    begin
      if Values[i].ExtremeOutliers <> nil then
        Values[i].ExtremeOutliers.Free;
      if Values[i].MildOutliers <> nil then
        Values[i].MildOutliers.Free;
    end;

  Players := aPlayers;
  Mode := aMode;

  Count := Length(Players);

  minV := 1000000000;
  maxV := -1000000000;

  SetLength(Values, Count);
  SetLength(BrightColors, Count);
  SetLength(DarkColors, Count);

  if Count = 0 then
    Exit;

  for i := 0 to Count - 1 do
  begin
    Values[i] := EmptyValue;
    BrightColors[i] := BlendScreen(Players[i].Color, TColor($DDDDDD));
    DarkColors[i] := BlendMultiply(Players[i].Color, TColor($777777));
  end;
end;

end.
