// * size
// # dropbox

unit IniObject;

interface

uses
  math, syncobjs, mask, crypt, forms, IniFile, Classes, Dialogs, SysUtils,
  Controls, ComCtrls, StdCtrls, FileCtrl, ExtCtrls, Buttons, Windows, StrUtils;

type
  IniGroup = class(TObject)
  public
    Name: string;
    Lines: TStringlist;
    Inited: boolean;
    Changed: boolean;
    constructor Create(n: string);
    destructor Destroy; override;
    procedure Clear;
    procedure SetData(n, d: string);
    function GetData(n: string): string;
    function Delete(n: string): Boolean;
    function Description(i: integer): string;
    function Data(i: integer): string;
    function Find(n: string): integer;
  end;

  IniFile = class(TObject)
  private
    Groups: TList;
  public
    Filename: string;
    encrypted: boolean;
    constructor Create(fn: string);
    destructor Destroy; override;
    procedure Clear;
    function Read: boolean;
    function UpdateUnchanged: boolean;
    function Save: boolean;
    function SaveChanges: boolean;
    function FindGroup(n: string): IniGroup;
    function DeleteGroup(n: string): boolean;
    function CreateGroup(n: string; CreateIfNotExists: Boolean = True): IniGroup;
    function Changed: boolean;
    procedure GroupNames(var List: tstringlist);
    procedure Rename(newName: string);
    procedure QuickSet(n, s: string); overload;
    procedure QuickSet(n: string; d: integer); overload;
    function QuickGet(n, default: string): string; overload;
    function QuickGet(n: string; default: integer): integer; overload;
    function Count: Integer;
  end;

  IniControlData = class(TObject)
  private
    ControlDescriptions: Tstringlist;
    ControlNames: TStringlist;
    FGroup: string;
    FIniFile: IniFile;
    FIniGroup: IniGroup;
    procedure AssignGroup;
  public
    ICDLoadControlSize: Boolean;
    ICDChangeAutoSizeColumns: Boolean;
    ICDLoadListViewWidth: Boolean;
    constructor Create(var aIniFile: IniFile; GroupName: string);
    destructor Destroy; override;
    procedure ReadControlList(ControlList: array of string); overload;
    procedure ReadControlList(CL: string); overload;
    procedure SaveControlData(Form: TWinControl);
    procedure LoadControlData(Form: TWinControl);
    procedure SaveCoords(Form: TWinControl);
    function LoadCoords(Form: TWinControl): boolean;
    function Changed: boolean;
  end;

type
  NewIniFile = IniFile;

implementation

{ IniGroup }

procedure IniGroup.Clear;
begin
  Lines.clear;
end;

function IniGroup.Data(i: integer): string;
var
  x: integer;
begin
  x := pos('=', lines[i]);
  if x > 0 then
    result := copy(Lines[i], x + 1, 65536)
  else
    result := '';
end;

function IniGroup.Description(i: integer): string;
var
  x: integer;
begin
  x := pos('=', lines[i]);
  if x > 0 then
    result := copy(Lines[i], 1, x - 1)
  else
    result := '';
end;

function IniGroup.Find(n: string): integer;
var
  i: integer;
begin
  for i := 0 to lines.count - 1 do
    if uppercase(n) = uppercase(description(i)) then
    begin
      result := i;
      exit;
    end;
  result := -1;
end;

function IniGroup.GetData(n: string): string;
var
  i: integer;
begin
  i := Find(n);
  if i >= 0 then
    result := Data(i)
  else
    result := 'not found';
end;

constructor IniGroup.Create(n: string);
begin
  lines := tstringlist.create;
  inited := true;
  name := n;
end;

destructor IniGroup.Destroy;
begin
  if not inited then
    exit;
  lines.free;
  name := '';
  inited := false;
  inherited;
end;

procedure IniGroup.SetData(n, d: string);
var
  i: integer;
begin
  i := find(n);
  if i > -1 then
    lines[i] := n + '=' + d
  else
    lines.Add(n + '=' + d);
  changed := true;
end;

function IniGroup.Delete(n: string): boolean;
var
  i: integer;
begin
  i := find(n);
  if i > -1 then
  begin
    lines.delete(i);
    result := true;
  end
  else
    result := false;
end;

{ IniFile }

function IniFile.Changed: boolean;
var
  i: integer;
begin
  result := false;
  for i := 0 to Count - 1 do
    result := result or (IniGroup(groups[i]).changed);
end;

procedure IniFile.Clear;
var
  i: integer;
  g: IniGroup;
begin
  for i := 0 to Count - 1 do
  begin
    g := groups[i];
    g.Free;
  end;
  groups.Clear;
end;

constructor IniFile.Create(fn: string);
begin
  groups := TList.Create;
  filename := fn;
  encrypted := false;
end;

function IniFile.CreateGroup(n: string; CreateIfNotExists: Boolean = True): IniGroup;
var
  p: IniGroup;
begin
  p := FindGroup(n);
  if (p = nil) and CreateIfNotExists then
  begin
    p := IniGroup.Create(n);
    p.Changed := true;
    Groups.Add(p);
  end;
  result := p;
end;

function IniFile.DeleteGroup(n: string): boolean;
var
  g: IniGroup;
begin
  g := FindGroup(n);
  if assigned(g) then
  begin
    g.Free;
    Groups.Remove(g);
    result := true;
  end else
    result := false;
end;

destructor IniFile.Destroy;
begin
  Clear;
  Groups.Free;
  inherited;
end;

function IniFile.FindGroup(n: string): IniGroup;
var
  p: IniGroup;
  i: integer;
begin
  for i := 0 to Count - 1 do
  begin
    p := groups[i];
    if uppercase(p.name) = uppercase(n) then
    begin
      result := p;
      exit;
    end;
  end;
  result := nil;
end;

function IniFile.Read: boolean;
var
  g: inigroup;
  a: string;
  i: integer;
  t: tstringlist;
begin
  Clear;
  result := false;
  if Fileexists(filename) then
  begin
    t := tstringlist.create;
    try
      t.LoadFromFile(Filename);
      if Encrypted then
        Decryptstringlist(t, 2012);
      i := 0;
      while i < t.count do
      begin
        a := t[i];
        inc(i);
        if (length(a) > 2) and (a[1] = '[') then
        begin
          g := CreateGroup(copy(a, 2, length(a) - 2));
          g.Changed := False;
          while i < t.count do
          begin
            if t[i] = '' then
              break
            else
              g.Lines.Add(t[i]);
            inc(i);
          end;
        end;
      end;
      result := true;
    except
      result := false;
    end;
    t.free;
  end;
end;

function IniFile.UpdateUnchanged: boolean;
var
  g: inigroup;
  a: string;
  z: integer;
  t: tstringlist;
  neu: boolean;
begin
  t := TStringList.create;
  try
    z := 0;
    t.LoadFromFile(filename);
    if Encrypted then
      Decryptstringlist(t, 2012);
    while z <> t.count do
    begin
      a := t[z];
      inc(z);
      if (length(a) > 2) and (a[1] = '[') then
      begin
        g := FindGroup(copy(a, 2, length(a) - 2));
        if g = nil then
        begin
          neu := true;
          g := CreateGroup(copy(a, 2, length(a) - 2));
          g.Changed := false;
        end else
          neu := false;
        if neu or (g.Changed = false) then
        begin
          g.Clear;
          while z <> t.Count do
          begin
            inc(z);
            if t[z] = '' then
              break
            else
              g.lines.Add(t[z]);
          end;
        end;
      end;
    end;
    t.free;
    result := true;
  except
    t.free;
    result := false;
  end;
end;

procedure IniFile.Rename(newName: string);
begin
  Renamefile(filename, newname);
  filename := newname;
end;

function IniFile.Save: boolean;
var
  t: tstringlist;
  j, i: integer;
begin
  t := tstringlist.create;
  try
    for i := 0 to Count - 1 do
    begin
      t.add('[' + IniGroup(groups[i]).name + ']');
      for j := 0 to IniGroup(groups[i]).lines.Count - 1 do
        t.add(IniGroup(groups[i]).lines[j]);
      t.add('');
      IniGroup(groups[i]).changed := false;
    end;
    if encrypted then
      encryptstringlist(t, 2012);
    t.SaveToFile(filename);
    t.free;
    result := true;
  except
    t.free;
    result := false;
  end;
end;

function IniFile.SaveChanges: boolean;
begin
  result := Updateunchanged and save;
end;

procedure IniFile.GroupNames(var list: TStringList);
var
  i: integer;
begin
  for i := 0 to Count - 1  do
    list.add(IniGroup(Groups[i]).name);
end;

function IniFile.QuickGet(n, default: string): string;
var
  g: IniGroup;
begin
  g := FindGroup('Misc');
  if g = nil then
    result := default
  else
    result := ifthen(g.GetData(n)='not found', default, g.getData(n));
end;

procedure IniFile.QuickSet(n, s: string);
var
  g: IniGroup;
begin
  g := FindGroup('Misc');
  if g = nil then
    g := CreateGroup('Misc');
  g.SetData(n, s);
end;

function IniFile.QuickGet(n: string; default: integer): integer;
begin
  Result := StrToInt(QuickGet(n, IntToStr(default)));
end;

procedure IniFile.QuickSet(n: string; d: integer);
begin
  QuickSet(n, IntToStr(d));
end;

function IniFile.Count: Integer;
begin
  result := Groups.Count;
end;

{ IniControlData }

procedure IniControlData.AssignGroup;
begin
  FIniGroup := FIniFile.CreateGroup(FGroup)
end;

function IniControlData.Changed: boolean;
begin
  result := FIniGroup.changed;
end;

constructor IniControlData.Create(var aIniFile: IniFile; Groupname: String);
begin
  if aIniFile = nil then
    raise(Exception.Create('IniControlData.Create' + cr +
      'aIniFile zeigt auf NIL'));
  ControlNames := TStringlist.create;
  ControlDescriptions := Tstringlist.create;
  FGroup := Groupname;
  FIniFile := aInifile;
  ICDLoadControlSize := True;
  ICDChangeAutoSizeColumns := True;
  ICDLoadListViewWidth := True;
end;

destructor IniControlData.Destroy;
begin
  ControlNames.Free;
  ControlDescriptions.Free;
  inherited;
end;

procedure IniControlData.LoadControlData(Form: TWinControl);
var
  m, x, u, n, i, j: integer;
  c: TControl;
  b: boolean;
  count, count2: integer;
  a: string;
  l: tlistitem;
  t: TStringlist;
begin
  AssignGroup;
  count2 := 0;
  t := Tstringlist.create;
  for i := 0 to (Form.ControlCount) - 1 do
  begin
    c := Form.Controls[i];
    if c is twincontrol then
      loadcontroldata(c as twincontrol);
    for n := 0 to ControlNames.count - 1 do
      if lowercase(c.name) = lowercase(ControlNames[n]) then
        if n > -1 then
        begin
          a := FIniGroup.GetData(ControlDescriptions[n]);
          if a = 'not found' then
            a := ''
          else
          begin
            if (pos('*', ControlDescriptions[n]) > 0) and (a <> '') then
            begin
              if ICDloadcontrolsize then
              begin
                (c as TControl).Width := strtoint(teilstring(a, 1));
                (c as TControl).Height := strtoint(teilstring(a, 2));
              end;
              if c is TPanel then
              begin
                (c as TControl).Width := strtoint(teilstring(a, 1));
                (c as TControl).Height := strtoint(teilstring(a, 2));
              end;
              if c is TListView then
                for j := 0 to (c as TListview).Columns.Count - 1 do
                begin
                  b := (c as TListView).Columns[j].AutoSize and
                    ICDChangeAutoSizeColumns;
                  if not b then
                    (c as TListView).Columns[j].Width := strtoint(teilstring(a, 3
                      + j))
                  else
                    (c as TListView).Columns[j].MaxWidth :=
                      strtoint(teilstring(a, 3 + j));
                end;
            end
            else if (pos('#', ControlDescriptions[n]) > 0) and (a <> '') then
            begin
              if c is TComboBox then
                (c as tcombobox).items.assign(iniaslist(a));
            end
            else
            begin
              if c is TListView then
              begin
                (c as TListView).items.BeginUpdate;
                (c as TListView).items.clear;
                x := 1;
                count := strtoint(teilstring(a, x));
                for u := 0 to count - 1 do
                begin
                  l := (c as TListview).Items.Add;
                  inc(x);
                  l.Caption := teilstring(a, x);
                  inc(x);
                  count2 := strtoint(teilstring(a, x));
                  t.clear;
                  for m := 1 to count2 do
                    t.add(teilstring(a, x + m));
                  l.subitems.assign(t);
                  x := x + count2;
                end;
                inc(x);
                if count = 0 then
                  count2 := iniaslist(a).Count - 2;
                if ICDLoadListViewWidth then
                  for m := 0 to min(count2, (c as TListView).Columns.count - 1)
                    do
                    (c as TListView).Columns[m].width := strtoint(teilstring(a, x
                      + m));
                (c as TListView).items.EndUpdate;
              end;
              if c is TDriveComboBox then
                if a <> '' then
                  (c as tdrivecombobox).drive := a[1];
              if c is TDirectorylistbox then
                if a <> '' then
                  (c as tdirectorylistbox).directory := a;
              if c is TPanel then
                if a <> '' then
                begin
                  (c as TPanel).Width := strtoint(teilstring(a, 1));
                  (c as TPanel).Height := strtoint(teilstring(a, 2));
                end;
              if c is TEdit then
                (c as Tedit).Text := a;
              if c is TMaskEdit then
                (c as Tmaskedit).Text := a;
              if c is TMemo then
                (c as TMemo).Lines.Assign(iniaslist(a));
              if c is TCheckbox then
                (C as TCheckbox).checked := IniAsBoolean(a);
              if c is TButton then
                (c as TButton).Caption := a;
              if c is TLabel then
                (c as TLabel).Caption := a;
              if c is TListBox then
                (c as TListbox).items.assign(iniaslist(a));
              if c is TTreeView then
                inistringastree(a, (c as TTreeView).items);
              if c is TComboBox then
              begin
                if (c as TCombobox).Style = csDropDown then
                  (c as TCombobox).text := a;
                if (c as TComboBox).Style = csDropDownList then
                  (c as TCombobox).itemindex := (c as
                    Tcombobox).items.indexof(a);
              end;
              if c is TRadioButton then
                (c as TRadioButton).Checked := iniasboolean(a);
              if c is TSpeedButton then
                (c as TSpeedButton).Down := iniasboolean(a);
              if c is TTrackBar then
                (c as TTrackBar).Position := StrToInt(a);
            end; // Ende a enthlt kein *
          end;
        end;
  end;
  T.Free;
end;

procedure IniControlData.SaveCoords(Form: TWinControl);
begin
  if (form as tform).windowstate <> wsnormal then
    exit;
  assigngroup;
  finigroup.SetData(form.name, inttostr(form.top) + '|' + inttostr(form.left) +
    '|' + inttostr(form.height) + '|' + inttostr(form.width));
end;

procedure IniControlData.ReadControlList(ControlList: array of string);
var
  i: integer;
  a: string;
begin
  a := Controllist[0];
  for i := 1 to high(Controllist) do
  begin
    a := a + ',' + Controllist[i];
  end;
  ReadControllist(a);
end;

procedure IniControlData.ReadControlList(CL: string);
var
  p: integer;
  a, b: string;
begin
  controlnames.Clear;
  controldescriptions.Clear;
  repeat
    p := pos(',', cl);
    if p > 0 then
    begin
      a := copy(cl, 1, p - 1);
      delete(cl, 1, p);
    end
    else
      a := cl;
    if not split(a, a, b, '|') then
      raise(Exception.Create('INIOBJECT: Illegal Controllist.'));
    ControlNames.Add(a);
    ControlDescriptions.Add(b);
  until p = 0;
end;

procedure IniControlData.SaveControlData(Form: TWinControl);
var
  u, n, i, j: integer;
  c: TControl;
  a: string;
  t: TStringlist;
  l: tListItem;
  cs: tcriticalsection;
begin
  cs := tcriticalsection.create;
  cs.enter;
  AssignGroup;
  t := Tstringlist.create;
  for i := 0 to (Form.ControlCount) - 1 do
  begin
    c := Form.Controls[i];
    if (c is twincontrol) then
      savecontroldata(c as twincontrol);
    for n := 0 to ControlNames.count - 1 do
      if lowercase(c.name) = lowercase(ControlNames[n]) then
      begin
        if pos('*', ControlDescriptions[n]) > 0 then
        begin
          a := inttostr((c as TControl).Width) + '|' + inttostr((c as
            TControl).Height);
          if c is TListView then
            for j := 0 to (c as Tlistview).columns.count - 1 do
              a := a + '|' + inttostr((c as tlistview).columns[j].width);
        end
        else if pos('#', ControlDescriptions[n]) > 0 then
        begin
          if c is TComboBox then
          begin
            t.assign((c as tcombobox).items);
            a := iniasstring(t);
          end;
        end
        else
        begin
          if c is TListView then
          begin
            a := inttostr((c as TListview).items.count);
            for u := 0 to (c as TListview).items.count - 1 do
            begin
              l := (c as TListView).Items[u];
              t.assign(l.subitems);
              if l.subitems.count > 0 then
                a := a + '|' + l.Caption + '|' + inttostr(l.subitems.count) + '|'
                  + iniasstring(t)
              else
                a := a + '|' + l.Caption + '|0';
            end;
            for u := 0 to (c as TListView).Columns.Count - 1 do
              a := a + '|' + inttostr((C as TListView).Columns[u].Width);
          end;
          if c is TDirectoryListbox then
            a := (c as TDirectoryListBox).Directory;
          if c is TDriveComboBox then
            a := (c as TDriveCombobox).Drive;
          if c is TEdit then
            a := (c as Tedit).Text;
          if c is TMaskEdit then
            a := (c as Tmaskedit).Text;
          if c is TPanel then
            a := inttostr((c as TPanel).Width) + '|' + inttostr((c as
              TPanel).Height);
          if c is TMemo then
          begin
            t.assign((c as TMemo).Lines);
            a := iniasstring(t);
          end;
          if c is TCheckbox then
            a := iniasvalue((C as TCheckbox).checked);
          if c is TButton then
            a := (c as TButton).Caption;
          if c is TLabel then
            a := (c as TLabel).Caption;
          if c is TTreeView then
            a := initreeasstring((c as ttreeview).items);
          if c is TListBox then
          begin
            t.assign((c as TListbox).items);
            a := iniasstring(t);
          end;
          if c is TComboBox then
            a := (c as TCombobox).text;
          if c is TRadioButton then
            a := iniasvalue((c as TRadioButton).Checked);
          if c is TSpeedbutton then
            a := iniasvalue((c as TSpeedbutton).Down);
          if c is TTrackBar then
            a := IntToStr((c as TTrackBar).Position);
        end;
        // derung: damit das Speichern leerer TEdits geht
        if {(a<>'') and }(a <> Finigroup.GetData(ControlDescriptions[n])) then
          FIniGroup.SetData(ControlDescriptions[n], a);
      end;
  end;
  cs.leave;
  cs.free;
  T.Free;
end;

function IniControlData.LoadCoords(Form: TWinControl): boolean;
var
  a: string;
begin
  assigngroup;
  a := finigroup.GetData(form.name);
  if a <> 'not found' then
  begin
    form.Top := strtoint(teilstring(a, 1));
    form.Left := strtoint(teilstring(a, 2));
    form.Height := strtoint(teilstring(a, 3));
    form.Width := strtoint(teilstring(a, 4));
    result := true;
  end
  else
  begin
    form.Left := screen.Width div 2 - form.Width div 2;
    form.top := Screen.Height div 2 - form.height div 2;
    result := false;
  end;
end;

end.

