unit Resources;

interface

uses Classes, SysUtils;

const
  SIG = 'RES';

type

  TResource = class(TObject)
  public
    Stream: TMemoryStream;
    Name: String;
    constructor Create;
    destructor Destroy; override;
  end;

  TResources = class(TObject)
  public
    Streams: TList;
    constructor Create;
    destructor Destroy; override;
    procedure LoadFromFile(FileName: String);
    procedure SaveToFile(FileName: String);
    procedure Clear;
    function GetResource(Name: String): TMemoryStream; overload;
    function GetResource(id: Integer): TMemoryStream; overload;
    function GetRawResource(id: Integer): TResource;
    function AddResource(Name: String): Integer;
    procedure DeleteResoure(id: Integer);
  end;

  TResourceReader = class(TObject)
  private
    f: TFileStream;
    m: TMemoryStream;
  public
    constructor Create(FileName: String);
    destructor Destroy; override;
    procedure Seek(NameofResource: String); overload;
    procedure Seek(id: Integer); overload;
    function Stream: TStream;
  end;


implementation

{ TResources }

function TResources.AddResource(Name: String): Integer;
var
  r: TResource;
begin
  r := TResource.Create;
  r.Name := Name;
  result := Streams.Add(r);
end;

procedure TResources.Clear;
var i: Integer;
begin
  for i := 0 to Streams.Count - 1 do
    TResource(Streams[i]).Free;
  Streams.Clear;
end;

constructor TResources.Create;
begin
  Streams := TList.Create;
end;

procedure TResources.DeleteResoure(id: Integer);
begin
  TResource(Streams[id]).Free;
  Streams.Delete(id);
end;

destructor TResources.Destroy;
begin
  Clear;
  Streams.Free;
  inherited;
end;

function TResources.GetResource(Name: String): TMemoryStream;
var
  i: Integer;
begin
  for i := 0 to Streams.Count - 1 do
    if Name = TResource(Streams[i]).Name then
    begin
      result := TResource(Streams[i]).Stream;
      result.Position := 0;
      exit;
    end;
  result := NIL;
end;

function TResources.GetResource(id: Integer): TMemoryStream;
begin
  if id < Streams.Count then
  begin
    result := TResource(Streams[id]).Stream;
    result.Position := 0;
  end else
    result := NIL;
end;

function TResources.GetRawResource(id: Integer): TResource;
begin
  if id < Streams.Count then
  begin
    result := TResource(Streams[id]);
  end else
    result := NIL;
end;

procedure TResources.LoadFromFile(FileName: String);
var
  f: TFileStream;
  c, i: Integer;
  l: Cardinal;
  s: array of Int64;
  r: TResource;
  name: PChar;
  signature: Array[0..2] of Char;
begin
  f := TFileStream.Create(FileName, fmOpenRead or fmShareDenyNone);
  try
    f.Read(signature, sizeof(signature));
    if signature <> SIG then
      raise Exception.Create('illegal resource file');
    f.Read(c, sizeof(c));
    setLength(s, c);
    Clear;
    for i := 0 to c - 1 do begin
      f.Read(l, sizeof(l));
      name := StrAlloc(l+1);
      fillchar(name^, l+1, 0);
      f.Read(name^, l);
      AddResource(name);
      StrDispose(name);
      f.Read(s[i], sizeof(Int64));
    end;
    for i := 0 to c - 1 do begin
      r := TResource(Streams[i]);
      r.Stream.CopyFrom(f, s[i]);
    end;
  finally
    f.Free;
    setlength(s, 0);
  end;
end;

procedure TResources.SaveToFile(FileName: String);
var
  f: TFileStream;
  c, i: Integer;
  l: Cardinal;
  s: Int64;
  r: TResource;
begin
  f := TFileStream.Create(FileName, fmCreate);
  c := Streams.Count;
  f.Write(SIG, length(SIG));
  f.Write(c, sizeof(c));
  for i := 0 to c - 1 do
  begin
    r := Streams[i];
    l := Length(r.Name);
    f.Write(l, sizeof(c));
    f.Write(r.Name[1], l);
    s := r.Stream.Size;
    f.Write(s, sizeof(s));
  end;
  for i := 0 to c - 1 do
  begin
    r := Streams[i];
    r.Stream.Position := 0;
    f.CopyFrom(r.Stream, r.Stream.Size);
  end;
  f.Free;
end;

{ TResource }

constructor TResource.Create;
begin
  Stream := TMemoryStream.Create;
  Name := '';
end;

destructor TResource.Destroy;
begin
  Stream.Free;
  inherited;
end;

{ TResourceReader }

constructor TResourceReader.Create(FileName: String);
begin
  f := TFileStream.Create(FileName, fmOpenRead or fmShareDenyNone);
  m := TMemoryStream.Create;
end;

destructor TResourceReader.Destroy;
begin
  f.Free;
  m.Free;
  inherited;
end;

procedure TResourceReader.Seek(NameofResource: String);
var
  c, i: Integer;
  l: Cardinal;
  s: array of Int64;
  p: Integer;
  name: PChar;
  signature: Array[0..2] of Char;
begin
  f.Position := 0;
  p := -1;
  try
    f.Read(signature, sizeof(signature));
    if signature <> SIG then
      raise Exception.Create('illegal resource file');
    f.Read(c, sizeof(c));
    setLength(s, c);
    for i := 0 to c - 1 do begin
      // Check name
      f.Read(l, sizeof(l));
      name := StrAlloc(l+1);
      fillchar(name^, l+1, 0);
      f.Read(name^, l);
      if name = NameofResource then
        p := i;
      StrDispose(name);
      // Read size of resource
      f.Read(s[i], sizeof(Int64));
    end;
    m.SetSize(0);
    m.Position := 0;
    if p >= 0 then
    begin
      for i := 0 to p - 1 do
        f.Seek(s[i], soCurrent);
      m.CopyFrom(f, s[p]);
      m.Position := 0;
    end;
  finally
    setlength(s, 0);
  end;
end;

procedure TResourceReader.Seek(id: Integer);
var
  c, i: Integer;
  l: Cardinal;
  s: array of Int64;
  p: Integer;
  signature: Array[0..2] of Char;
begin
  f.Position := 0;
  p := -1;
  try
    f.Read(signature, sizeof(signature));
    if signature <> SIG then
      raise Exception.Create('illegal resource file');
    f.Read(c, sizeof(c));
    setLength(s, c);
    for i := 0 to c - 1 do begin
      // Check name
      f.Read(l, sizeof(l));
      f.Seek(l, soCurrent);
      if id = i then
        p := i;
      // Read size of resource
      f.Read(s[i], sizeof(Int64));
    end;
    m.SetSize(0);
    m.Position := 0;
    if p >= 0 then
    begin
      for i := 0 to p - 1 do
        f.Seek(s[i], soCurrent);
      m.CopyFrom(f, s[p]);
      m.Position := 0;
    end;
  finally
    setlength(s, 0);
  end;
end;

function TResourceReader.Stream: TStream;
begin
  result := m;
end;

end.
