unit ULatency;

interface

uses
  Classes, SyncObjs, SysUtils, Math, Windows, WC3Types;

const
  fFIN = 1;
  fSYN = 2;
  fRST = 4;
  fPSH = 8;
  fACK = 16;
  fURG = 32;

type
  TLatencyPacketInfo = record
    srcIP: Cardinal;
    dstIP: Cardinal;
    srcPort: Word;
    dstPort: Word;
    flags: Word;
    tcpSize: Word;
    SEQ: Cardinal;
    ACK: Cardinal;
    time: Int64;
    useInLobby: Boolean;
  end;
  PLatencyPacketInfo = ^TLatencyPacketInfo;

  TLatencyConnection = class(TObject)
  public
    ip: Cardinal;
    port: Word;
    times: TList;
    maxCount: Integer;
    hsLatency: Cardinal;
    constructor Create;
    destructor Destroy; override;
    procedure AddTime(uTime: Cardinal);
    function GetAverage: Cardinal;
    function GetStandardDeviation: Cardinal;
  end;

  TLatencyHandler = class(TThread)
  private
    MaxCount: Integer;
    FEnabled: Boolean;
    procedure SetEnabled(const Value: Boolean);
  protected
    procedure Execute; override;
  public
    PacketList: TThreadList;
    Connections: TThreadList;
    GamePort: Integer;
    constructor Create;
    destructor Destroy; override;
    procedure AddPacket(srcIP: Cardinal; dstIP: Cardinal; srcPort: Word; dstPort: Word; Flags: Word; ACK: Cardinal; SEQ: Cardinal; TCPSize: Word; time: Int64; useInLobby: Boolean = false);
    procedure AddLatency(remoteIP: Cardinal; remotePort: Word; latency: Integer);
    function  GetLatency(remoteIP: Cardinal; remotePort: Word): Cardinal;
    function  GetStandardDeviation(remoteIP: Cardinal; remotePort: Word): Cardinal;
    procedure ClearConnectionList;
    procedure SetMaxCount(maxCount: Integer);
    procedure SetHSLatency(remoteIP: Cardinal; remotePort: Word; latency: Cardinal);
    function  GetHSLatency(remoteIP: Cardinal; remotePort: Word): Cardinal;
    function  GetValues(remoteIP: Cardinal; remotePort: Word): String;
    procedure GetLastValues(remoteIP: Cardinal; remotePort: Word; n: Integer; const targetList: TList);
    property Enabled: Boolean read FEnabled write SetEnabled; 
  end;

implementation

uses Unit1;

{ TLatencyHandler }

procedure TLatencyHandler.AddLatency(remoteIP: Cardinal; remotePort: Word;
  latency: Integer);
var
  i: Integer;
  c: TLatencyConnection;
begin
  if latency <= 0 then exit;
  c := nil;
  try
    with Connections.LockList do
    begin
      for i := 0 to Count - 1 do
        with TLatencyConnection(Items[i]) do
          if (ip = remoteIP) and (port = remotePort) then
          begin
            c := TLatencyConnection(Items[i]);
            break;
          end;
      if c = nil then
      begin
        c := TLatencyConnection.Create;
        c.maxCount := self.MaxCount;
        c.ip := remoteIP;
        c.port := remotePort;
        Add(c);
      end;
      c.AddTime(latency);
    end;
  finally
    Connections.UnlockList;
  end;
end;

procedure TLatencyHandler.AddPacket(srcIP: Cardinal; dstIP: Cardinal;
  srcPort, dstPort, Flags: Word; ACK, SEQ: Cardinal; TCPSize: Word; time: Int64;
  useInLobby: Boolean = false);
var
  p: PLatencyPacketInfo;
  ba: PByteArray;
begin
  if not Enabled or not assigned(PacketList) then
    exit;
  if (srcPort <> GamePort) and (dstPort <> GamePort) then
    exit;

  // add packet to packet list
  new(p);
  p^.srcIP := srcIP;
  p^.dstIP := dstIP;
  p^.srcPort := srcPort;
  p^.dstPort := dstPort;
  p^.flags := Swap(Flags);
  ba := @ACK;
  p^.ACK := (ba^[0] shl 24) or (ba^[1] shl 16) or (ba^[2] shl 8) or (ba^[3]);
  ba := @SEQ;
  p^.SEQ := (ba^[0] shl 24) or (ba^[1] shl 16) or (ba^[2] shl 8) or (ba^[3]);
  p^.tcpSize := TCPSize;
  p^.time := time;
  p^.useInLobby := useInLobby;
  PacketList.Add(p);
end;

procedure TLatencyHandler.ClearConnectionList;
var
  i: Integer;
begin
  try
    with Connections.LockList do
    begin
      for i := 0 to Count - 1 do
        TLatencyConnection(Items[i]).Free;
      Clear;
    end;
  finally
    Connections.UnlockList;
  end;
end;

constructor TLatencyHandler.Create;
begin
  Connections := TThreadList.Create;
  PacketList := TThreadList.Create;
  MaxCount := 100;
  Enabled := true;
  inherited Create(false);
end;

destructor TLatencyHandler.Destroy;
var
  i: Integer;
begin
  ClearConnectionList;
  Connections.Free;
  with PacketList.LockList do
  begin
    for i := 0 to Count - 1 do
      Dispose(Items[i]);
    Clear;
  end;
  PacketList.UnlockList;
  FreeAndNil(PacketList);
  inherited;
end;

procedure TLatencyHandler.Execute;
var
  i: Integer;
  j: Integer;
  ACKs, PSHs: TList;
  ACK, TMP, PSH: PLatencyPacketInfo;
  newestPacketTime: Int64;
begin
  ACKs := TList.Create;
  PSHs := TList.Create;
  while not Terminated do
  begin
    Sleep(500);
    if Enabled then
    try
      with PacketList.LockList do
      begin
        i := Count - 1;
        while i > -1 do
        begin
          ACK := Items[i];
          // found ACK directed to my PC ?
          if ((ACK^.flags and fACK) = fACK) and WC3Banlist_Main.InMyIP(ACK^.dstIP) then
          begin
            ACKs.Insert(0, ACK);
            for j := i - 1 downto 0 do
            begin
              TMP := items[j];
              // double ACK?
              if
                ((TMP^.flags and fACK) = fACK) and
                (TMP^.srcIP = ACK^.srcIP) and (TMP^.dstIP = ACK^.dstIP) and
                (TMP^.srcPort = ACK^.srcPort) and (TMP^.dstPort = ACK^.dstPort) and
                (TMP^.ACK = ACK^.ACK) {and (TMP^.SEQ = ACK^.SEQ)} then
                ACKs.Insert(0, TMP)
              // packet matching previously found ACK?
              else
              if   (TMP^.srcIP = ACK^.dstIP) and (TMP^.srcPort = ACK^.dstPort) and
                   (TMP^.dstIP = ACK^.srcIP) and (TMP^.dstPort = ACK^.srcPort) and
                   (TMP^.ACK = ACK^.SEQ) then
              begin
                if (((TMP^.flags and fPSH) = fPSH) and (ACK^.ACK = TMP^.SEQ + TMP^.tcpSize)) or
                   (((TMP^.flags and fSYN) = fSYN) and ((TMP^.flags and fACK) = fACK) and (ACK^.ACK = TMP^.SEQ + 1)) then
                  PSHs.Insert(0, TMP);
              end;
            end;
            // calc time if matching psh-packet found
            if (PSHs.Count > 0) then
            begin
              PSH := PSHs.First;
              ACK := ACKs.First;
              if (((PSH^.flags and fSYN) = fSYN) and ((PSH^.flags and fACK) = fACK)) or PSH^.useInLobby
              then
                SetHSLatency(ACK^.srcIP, ACK^.srcPort, (ACK^.time - PSH^.time) div 1000)
              else
                AddLatency(ACK^.srcIP, ACK^.srcPort, (ACK^.time - PSH^.time) div 1000);
            end;
            // clean up
            Dec(i, PSHs.Count + ACKs.Count);
            for j := 0 to ACKs.Count - 1 do
            begin
              Dispose(ACKs[j]);
              Remove(ACKs[j]);
            end;
            ACKs.Clear;
            for j := 0 to PSHs.Count - 1 do
            begin
              Dispose(PSHs[j]);
              Remove(PSHs[j]);
            end;
            PSHs.Clear;
          end else
            dec(i);
        end;
        // remove packets older than 3 sec
        i := Count;
        if (i > 0) then
        begin
          newestPacketTime := PLatencyPacketInfo(Last)^.time;
          while i > 0 do
            if (newestPacketTime - PLatencyPacketInfo(First)^.Time > 3000000) then
            begin
              Dispose(First);
              Delete(0);
              Dec(i);
            end else
              break;
        end;
      end;
    finally
      PacketList.UnlockList;
    end;
  end;
  ACKs.Free;
  PSHs.Free;
end;

function TLatencyHandler.GetHSLatency(remoteIP: Cardinal;
  remotePort: Word): Cardinal;
var
  i: Integer;
begin
  result := 0;
  try
    with Connections.LockList do
    begin
      i := Count - 1;
      while i > -1 do
        with TLatencyConnection(Items[i]) do
        begin
          if (remoteIP = ip) and (remotePort = port) then
          begin
            result := hsLatency;
            break;
          end;
          dec(i);
        end;
    end;
  finally
    Connections.UnlockList;
  end;
end;

procedure TLatencyHandler.GetLastValues(remoteIP: Cardinal;
  remotePort: Word; n: Integer; const targetList: TList);
var
  i: Integer;
  j: Integer;
begin
  try
    with Connections.LockList do
    begin
      i := Count - 1;
      while i > -1 do
        with TLatencyConnection(Items[i]) do
        begin
          if (remoteIP = ip) and (remotePort = port) then
          begin
            for j := times.Count - 1 downto max(0, times.Count - n) do
              targetList.Add(times[j]);
            break;
          end;
          dec(i);
        end;
    end;
  finally
    Connections.UnlockList;
  end;
end;

function TLatencyHandler.GetLatency(remoteIP: Cardinal;
  remotePort: Word): Cardinal;
var
  i: Integer;
begin
  result := 0;
  try
    with Connections.LockList do
    begin
      i := Count - 1;
      while i > -1 do
        with TLatencyConnection(Items[i]) do
        begin
          if (remoteIP = ip) and (remotePort = port) then
          begin
            result := GetAverage;
            break;
          end;
          dec(i);
        end;
    end;
  finally
    Connections.UnlockList;
  end;
end;

function TLatencyHandler.GetStandardDeviation(remoteIP: Cardinal;
  remotePort: Word): Cardinal;
var
  i: Integer;
begin
  result := 0;
  with Connections.LockList do
  begin
    i := Count - 1;
    while i > -1 do
      with TLatencyConnection(Items[i]) do
      begin
        if (remoteIP = ip) and (remotePort = port) then
        begin
          result := GetStandardDeviation;
          break;
        end;
        dec(i);
      end;
  end;
  Connections.UnlockList;
end;

function TLatencyHandler.GetValues(remoteIP: Cardinal;
  remotePort: Word): String;
var
  i: Integer;
  c: TLatencyConnection;
  delimiter: String;
begin
  c := nil;
  result := '';
  with Connections.LockList do
  begin
    for i := 0 to Count - 1 do
    begin
      with TLatencyConnection(Items[i]) do
      begin
        if (ip = remoteIP) and (port = remotePort) then
        begin
          c := TLatencyConnection(Items[i]);
          break;
        end;
      end;
    end;
    if c <> nil then
    begin
      delimiter := '';
      with c.times do
      for i := 0 to Count - 1 do
      begin
        result := result + delimiter + IntToStr(Cardinal(Items[i]));
        delimiter := ' ';
      end;
    end;
  end;
  Connections.UnlockList;
end;

procedure TLatencyHandler.SetEnabled(const Value: Boolean);
begin
  FEnabled := Value;
  if not Value then
    ClearConnectionList;
end;

procedure TLatencyHandler.SetHSLatency(remoteIP: Cardinal;
  remotePort: Word; latency: Cardinal);
var
  i: Integer;
  c: TLatencyConnection;
begin
  c := nil;
  try
    with Connections.LockList do
    begin
      for i := 0 to Count - 1 do
      begin
        with TLatencyConnection(Items[i]) do
        begin
          if (ip = remoteIP) and (port = remotePort) then
          begin
            c := TLatencyConnection(Items[i]);
            break;
          end;
        end;
      end;
      if c = nil then
      begin
        c := TLatencyConnection.Create;
        c.maxCount := self.MaxCount;
        c.ip := remoteIP;
        c.port := remotePort;
        add(c);
      end;
      c.hsLatency := latency;
    end;
  finally
    Connections.UnlockList;
  end;
end;

procedure TLatencyHandler.SetMaxCount(maxCount: Integer);
var
  i: Integer;
begin
  self.MaxCount := maxCount;
  try
    with Connections.LockList do
    begin
      for i := 0 to Count - 1 do
        TLatencyConnection(Items[i]).maxCount := maxCount;
    end;
  finally
    Connections.UnlockList;
  end;
end;

{ TLatencyConnection }

procedure TLatencyConnection.AddTime(uTime: Cardinal);
begin
  times.Add(Pointer(uTime));
  while times.Count > maxCount do
    times.Delete(0);
end;

constructor TLatencyConnection.Create;
begin
  times := TList.Create;
  hsLatency := 0;
  ip := 0;
  port := 0;
  maxCount := 0;
end;

destructor TLatencyConnection.Destroy;
begin
  times.Free;
  inherited;
end;

function TLatencyConnection.GetAverage: Cardinal;
var
  i: Integer;
begin
  result := 0;
  if times.Count > 0 then
  begin
    for i := 0 to times.Count - 1 do
      result := result + Cardinal(times[i]);
    result := result div Cardinal(times.Count);
  end;
end;

function TLatencyConnection.GetStandardDeviation: Cardinal;
var
  sum: Cardinal;
  avg: Integer;
  i: Integer;
  cnt: Integer;
begin
  result := 0;
  with times do
  if Count > 1 then
  begin
    cnt := Count - 1;
    avg := GetAverage;
    sum := 0;
    for i := 0 to Count - 1 do
      inc(sum, Sqr(avg - Integer(items[i])) div cnt);
    result := Round(Sqrt(sum));
  end;
end;

end.
