unit UListContainer;

interface

  uses XMLIntf, XMLDoc, Classes, Forms, Variants;

  type
    PBan = ^TBan;
    TBan = record
      nick: String;
      deleted: Boolean;
      comment: String;
      lastChange: TDateTime;
    end;

    PInfo = ^TInfo;
    TInfo = record
      nick: String;
      deleted: Boolean;
      comment: String;
      lastChange: TDateTime;
    end;

    PScore = ^TScore;
    TScore = record
      nick: String;
      gamesPlayed: Integer;
      gamesRated: Integer;
      points: Double;
      lastSeen: TDateTime;
    end;

    PGame = ^TGame;
    TGame = record
      id: String;
      ownNickName: String;
      realm: String;
      players: String;
      givenPoints: String;
      screenShots: String;
      mapName: String;
      gameName: String;
      replay: String;
      duration: String;
      gameStart: TDateTime;
      comment: String;
      host: String;
    end;

    TEntryList = class(TObject)
    public
      list: TList;
      listType: Integer;
      constructor Create(listType: Integer);
      destructor Destroy; override;
      procedure Clear;
      procedure SortByNickName;
      function FindEntry(nick: String; var index: Integer): Boolean;
      function AddEntry(p: Pointer) : Integer;
      procedure ImportFromXML(FileName: String);
    end;

    TListContainer = class(TObject)
    private
      owner: TForm;
    public
      banlist: TEntryList;
      infolist: TEntryList;
      scorelist: TEntryList;
      gamelist: TEntryList;
      constructor Create(AOwner: TForm);
      destructor Destroy; override;
      procedure SaveToFile(FileName: String);
      procedure LoadFromFile(FileName: String);
    end;

    procedure CreateBanNodes(l: TList; nbanlist: IXMLNode);
    procedure CreateInfoNodes(l: TList; ninfolist: IXMLNode);
    procedure ExportBans(l: TList; FileName: String; asXML: boolean);
    procedure ExportInfos(l: TList; FileName: String; asXML: boolean);


implementation

uses SysUtils;

const
  LTBANLIST = 0;
  LTSCORELIST = 1;
  LTINFOLIST = 2;
  LTGAMELIST = 3;
  LCVERSION = '1.0';

function TimeToStr(t: TDateTime): String;
var
  h, m, s, ms: Word;
begin
  DecodeTime(t, h, m, s, ms);
  result := Format('%.2d:%.2d:%.2d', [h,m,s]);
end;

function DateToStr(t: TDateTime): String;
var
  y, m, d: Word;
begin
  DecodeDate(t, y, m, d);
  result := Format('%.4d-%.2d-%.2d', [y, m, d]);
end;

function StrToTime(a: String): TDateTime;
var
  h, m, s: Word;
begin
  h := StrToInt(Copy(a, 1, 2));
  m := StrToInt(Copy(a, 4, 2));
  s := StrToInt(Copy(a, 7, 2));
  result := EncodeTime(h, m, s, 0);
end;

function StrToDate(a: String): TDateTime;
var
  y, m, d: Word;
begin
  y := StrToInt(Copy(a, 1, 4));
  m := StrToInt(Copy(a, 6, 2));
  d := StrToInt(Copy(a, 9, 2));
  result := EncodeDate(y, m, d);
end;

function Default(NodeValue: OleVariant; Default: OleVariant): OleVariant;
begin
  if VarIsNull(NodeValue) then
    result := default
  else
    result := NodeValue;
end;

function GetListFromNodes(desc: String; node: IXMLNode): String;
var
  i: Integer;
  d: String;
begin
  d := '';
  result := '';
  for i := 0 to node.ChildNodes.Count - 1 do
  if node.ChildNodes[i].NodeName = desc then
  begin
    result := result + d + node.ChildNodes[i].NodeValue;
    d := '|';
  end;
end;

procedure CreateNodesFromList(desc: String; node: IXMLNode; values: String);
var
  s: String;
  p: Integer;
begin
  while values <> '' do
  begin
    p := pos('|', values);
    if p = 0 then
    begin
      s := values;
      values := '';
    end else
    begin
      s := copy(values, 1, p - 1);
      delete(values, 1, p);
    end;
    node.AddChild(desc).NodeValue := s;
  end;
end;

function CompareBansByNickName(p1, p2: Pointer): Integer;
var
  i1, i2: PBan;
begin
  i1 := p1;
  i2 := p2;
  result := CompareText(lowercase(i1^.nick), lowercase(i2^.nick));
end;

function CompareInfosByNickName(p1, p2: Pointer): Integer;
var
  i1, i2: PInfo;
begin
  i1 := p1;
  i2 := p2;
  result := CompareText(lowercase(i1^.nick), lowercase(i2^.nick));
end;

function CompareScoresByNickName(p1, p2: Pointer): Integer;
var
  i1, i2: PScore;
begin
  i1 := p1;
  i2 := p2;
  result := CompareText(lowercase(i1^.nick), lowercase(i2^.nick));
end;

function CompareGamesByID(p1, p2: Pointer): Integer;
var
  i1, i2: PGame;
begin
  i1 := p1;
  i2 := p2;
  result := CompareText(i1^.ID, i2^.ID);
end;

{ TEntryList }

function TEntryList.AddEntry(p: Pointer): Integer;
var
  i: Integer;
  found: boolean;
begin
  case listType of
    LTBANLIST: found := FindEntry(PBan(p)^.nick, i);
    LTSCORELIST: found := FindEntry(PScore(p)^.nick, i);
    LTINFOLIST: found := FindEntry(PInfo(p)^.nick, i);
    LTGAMELIST: found := FindEntry(PGame(p)^.id, i);
  else
    raise Exception.Create('AddEntry: unknown list type');
  end;
  if found then
  begin
    dispose(list[i]);
    list.Delete(i);
    list.Insert(i, p);
    result := i;
  end else
  begin
    list.Insert(i, p);
    result := i;
  end;
end;

procedure TEntryList.Clear;
var
  i: Integer;
begin
  for i := 0 to list.Count - 1 do
    Dispose(list[i]);
  list.Clear;
end;

constructor TEntryList.Create(listType: Integer);
begin
  self.listType := listType;
  list := TList.Create;
end;

destructor TEntryList.Destroy;
begin
  Clear;
  list.Free;
  inherited;
end;

function TEntryList.FindEntry(nick: String; var index: Integer): Boolean;
var
  i, j, m, c: Integer;
  found: boolean;
begin
  i := 0;
  j := list.Count - 1;
  found := false;
  nick := lowercase(nick);
  c := -1;
  m := 0;
  while (i <= j) do
  begin
    m := (i + j) div 2;
    case listType of
      LTBANLIST: c := CompareText(nick, lowercase(PBan(list[m])^.nick));
      LTSCORELIST: c := CompareText(nick, lowercase(PScore(list[m])^.nick));
      LTINFOLIST: c := CompareText(nick, lowercase(PInfo(list[m])^.nick));
      LTGAMELIST: c := CompareText(nick, PGame(list[m])^.id);
    else
      raise Exception.Create('FindEntry: unknown list type');
    end;
    if c < 0 then j := m - 1 else
    if c > 0 then i := m + 1 else
    begin
      found := true;
      break;
    end;
  end;
  result := found;
  if c < 0 then index := m else
  if c > 0 then index := m + 1 else
    index := m;
end;

procedure TEntryList.ImportFromXML(FileName: String);
var
  xml: IXMLDocument;
  i: Integer;
  pi: PInfo;
  pb: PBan;
  n, nbanlist, ninfolist: IXMLNode;
begin
  xml := LoadXMLDocument(FileName);
  if listType = LTINFOLIST then begin
    ninfolist := xml.ChildNodes['infolist'];
    for i := 0 to ninfolist.ChildNodes.Count - 1 do
    begin
      new(pi);
      n := ninfolist.ChildNodes[i];
      pi^.nick := n.Attributes['nick'];
      pi^.deleted := n.Attributes['deleted'];
      pi^.comment := Default(n.ChildNodes['comment'].NodeValue, '');
      pi^.lastChange := StrToDate(Default(n.ChildNodes['lastchange'].ChildNodes['date'].NodeValue, Date));
      pi^.lastChange := pi^.lastChange + StrToTime(Default(n.ChildNodes['lastchange'].ChildNodes['time'].NodeValue, Time));
      AddEntry(pi)
    end;
  end else if listType = LTBANLIST then begin
    nbanlist := xml.ChildNodes['banlist'];
    for i := 0 to nbanlist.ChildNodes.Count - 1 do
    begin
      new(pb);
      n := nbanlist.ChildNodes[i];
      pb^.nick := n.Attributes['nick'];
      pb^.deleted := n.Attributes['deleted'];
      pb^.comment := Default(n.ChildNodes['comment'].NodeValue, '');
      pb^.lastChange := StrToDate(Default(n.ChildNodes['lastchange'].ChildNodes['date'].NodeValue, Date));
      pb^.lastChange := pb^.lastChange + StrToTime(Default(n.ChildNodes['lastchange'].ChildNodes['time'].NodeValue, Time));
      AddEntry(pb);
    end;
  end;
  SortByNickName;
end;

procedure TEntryList.SortByNickName;
begin
  case listType of
   LTBANLIST: list.Sort(CompareBansByNickName);
   LTINFOLIST: list.Sort(CompareInfosByNickName);
   LTSCORELIST: list.Sort(CompareScoresByNickName);
   LTGAMELIST: list.Sort(CompareGamesByID);
  end;
end;

{ TListContainer }

constructor TListContainer.Create(AOwner: TForm);
begin
  owner := AOwner;
  banlist := TEntryList.Create(LTBANLIST);
  scorelist := TEntryList.Create(LTSCORELIST);
  infolist := TEntryList.Create(LTINFOLIST);
  gamelist := TEntryList.Create(LTGAMELIST);
end;

destructor TListContainer.Destroy;
begin
  banlist.Free;
  scorelist.Free;
  infolist.Free;
  gamelist.Free;
  inherited;
end;

procedure TListContainer.LoadFromFile(FileName: String);
var
  xml: IXMLDocument;
  database, nbanlist, ninfolist, nscorelist, ngamelist: IXMLnode;
  n: IXMLNode;
  i: Integer;
  pb: PBan;
  pi: PInfo;
  ps: PScore;
  pg: PGame;
begin
  xml := LoadXMLDocument(FileName);
  database := xml.DocumentElement;
  if (database.NodeName <> 'database') then
    raise EInOutError.Create('database node not found');
  banlist.Clear;
  infolist.Clear;
  scorelist.Clear;
  nbanlist := database.ChildNodes['banlist'];
  ninfolist := database.ChildNodes['infolist'];
  nscorelist := database.ChildNodes['scorelist'];
  ngamelist := database.ChildNodes['gamelist'];
  for i := 0 to nbanlist.ChildNodes.Count - 1 do
  begin
    new(pb);
    n := nbanlist.ChildNodes[i];
    pb^.nick := n.Attributes['nick'];
    pb^.deleted := n.Attributes['deleted'];
    pb^.comment := Default(n.ChildNodes['comment'].NodeValue, '');
    pb^.lastChange := StrToDate(Default(n.ChildNodes['lastchange'].ChildNodes['date'].NodeValue, Date));
    pb^.lastChange := pb^.lastChange + StrToTime(Default(n.ChildNodes['lastchange'].ChildNodes['time'].NodeValue, Time));
    banlist.list.Add(pb);
  end;
  for i := 0 to ninfolist.ChildNodes.Count - 1 do
  begin
    new(pi);
    n := ninfolist.ChildNodes[i];
    pi^.nick := n.Attributes['nick'];
    pi^.deleted := n.Attributes['deleted'];
    pi^.comment := Default(n.ChildNodes['comment'].NodeValue, '');
    pi^.lastChange := StrToDate(Default(n.ChildNodes['lastchange'].ChildNodes['date'].NodeValue, Date));
    pi^.lastChange := pi^.lastChange + StrToTime(Default(n.ChildNodes['lastchange'].ChildNodes['time'].NodeValue, Time));
    infolist.list.Add(pi);
  end;
  for i := 0 to nscorelist.ChildNodes.Count - 1 do
  begin
    new(ps);
    n := nscorelist.ChildNodes[i];
    ps^.nick := n.Attributes['nick'];
    ps^.gamesPlayed := Default(n.Attributes['gamesplayed'], 0);
    ps^.gamesRated := Default(n.Attributes['gamesrated'], 0);
    ps^.points := Default(n.Attributes['points'], 0);
    ps^.lastSeen := StrToDate(Default(n.ChildNodes['lastseen'].ChildNodes['date'].NodeValue, Date));
    ps^.lastSeen := ps^.lastSeen + StrToTime(Default(n.ChildNodes['lastseen'].ChildNodes['time'].NodeValue, Time));
    scorelist.list.Add(ps)
  end;
  for i := 0 to ngamelist.ChildNodes.Count - 1 do
  begin
    new(pg);
    n := ngamelist.ChildNodes[i];
    pg^.id := n.Attributes['id'];
    pg^.ownNickName := n.ChildNodes['ownnickname'].NodeValue;
    pg^.realm := n.ChildNodes['realm'].NodeValue;
    pg^.players := GetListFromNodes('player', n.ChildNodes['players']);
    pg^.givenPoints := GetListFromNodes('score', n.ChildNodes['givenscores']);
    pg^.screenShots := GetListFromNodes('screenshot', n.ChildNodes['screenshots']);
    pg^.mapName := Default(n.ChildNodes['mapname'].NodeValue, '');
    pg^.gameName := Default(n.ChildNodes['gamename'].NodeValue, '');
    pg^.replay := Default(n.ChildNodes['replay'].NodeValue, '');
    pg^.duration := Default(n.ChildNodes['duration'].NodeValue, 'unknown');
    pg^.gameStart := StrToDate(n.ChildNodes['gamestart'].ChildNodes['date'].NodeValue);
    pg^.gameStart := pg^.gameStart + StrToTime(n.ChildNodes['gamestart'].ChildNodes['time'].NodeValue);
    pg^.comment := Default(n.ChildNodes['comment'].NodeValue, '');
    pg^.host := Default(n.ChildNodes['host'].NodeValue, '');
    gamelist.list.Add(pg);
  end;
  banlist.SortByNickName;
  infolist.SortByNickName;
  scorelist.SortByNickName;
  gamelist.SortByNickName;
end;

procedure TListContainer.SaveToFile(FileName: String);
var
  xml: IXMLDocument;
  database, nbanlist, ninfolist, nscorelist, ngamelist: IXMLnode;
  n, m: IXMLNode;
  i: Integer;
  ps: PScore;
  pg: PGame;
begin
  xml := NewXMLDocument();
  xml.Encoding := 'UTF-8';
  xml.Options := xml.Options + [doNodeAutoIndent];
  xml.NodeIndentStr := ' ';
  database := xml.AddChild('database');
  database.Attributes['version'] := LCVERSION;
  nbanlist := database.AddChild('banlist');
  ninfolist := database.AddChild('infolist');
  nscorelist := database.AddChild('scorelist');
  ngamelist := database.AddChild('gamelist');
  // banlist nodes
  CreateBanNodes(banlist.list, nbanlist);
  // infolist nodes
  CreateInfoNodes(infolist.list, ninfolist);
  // scorelist nodes
  for i := 0 to scorelist.list.Count - 1 do
  begin
    ps := scorelist.list[i];
    n := nscorelist.AddChild('score');
    n.Attributes['nick'] := ps^.nick;
    n.Attributes['gamesplayed'] := ps^.gamesPlayed;
    n.Attributes['gamesrated'] := ps^.gamesRated;
    n.Attributes['points'] := ps^.points;
    m := n.AddChild('lastseen');
    m.AddChild('date').NodeValue := DateToStr(ps^.lastSeen);
    m.AddChild('time').NodeValue := TimeToStr(ps^.lastSeen);
  end;
  // gamelist nodes
  for i := 0 to gamelist.list.Count - 1 do
  begin
    pg := gamelist.list[i];
    n := ngamelist.AddChild('game');
    n.Attributes['id'] := pg^.id;
    n.AddChild('ownnickname').NodeValue := pg^.ownNickName;
    n.AddChild('realm').NodeValue := pg^.realm;
    CreateNodesFromList('player', n.AddChild('players'), pg^.players);
    CreateNodesFromList('score', n.AddChild('givenscores'), pg^.givenPoints);
    CreateNodesFromList('screenshot', n.AddChild('screenshots'), pg^.screenShots);
    n.AddChild('mapname').NodeValue := pg^.mapName;
    n.AddChild('gamename').NodeValue := pg^.gameName;
    n.AddChild('replay').NodeValue := pg^.replay;
    n.AddChild('duration').NodeValue := pg^.duration;
    m := n.AddChild('gamestart');
    m.AddChild('date').NodeValue := DateToStr(pg^.gameStart);
    m.AddChild('time').NodeValue := TimeToStr(pg^.gameStart);
    n.AddChild('comment').NodeValue := pg^.comment;
    n.AddChild('host').NodeValue := pg^.host;
  end;
  xml.SaveToFile(FileName);
end;

procedure CreateBanNodes(l: TList; nbanlist: IXMLNode);
var
  n, m: IXmlNode;
  pb: PBan;
  i: Integer;
begin
  for i := 0 to l.Count - 1 do
  begin
    pb := l[i];
    n := nbanlist.AddChild('ban');
    n.Attributes['nick'] := pb^.nick;
    n.Attributes['deleted'] := ord(pb^.deleted);
    n.AddChild('comment').NodeValue := pb^.comment;
    m := n.AddChild('lastchange');
    m.AddChild('date').NodeValue := DateToStr(pb^.lastChange);
    m.AddChild('time').NodeValue := TimeToStr(pb^.lastChange);
  end;
end;

procedure CreateInfoNodes(l: TList; ninfolist: IXMLNode);
var
  i: Integer;
  n, m: IXMLNode;
  pi: PInfo;
begin
  for i := 0 to l.Count - 1 do
  begin
    pi := l[i];
    n := ninfolist.AddChild('info');
    n.Attributes['nick'] := pi^.nick;
    n.Attributes['deleted'] := ord(pi^.deleted);
    n.AddChild('comment').NodeValue := pi^.comment;
    m := n.AddChild('lastchange');
    m.AddChild('date').NodeValue := DateToStr(pi^.lastChange);
    m.AddChild('time').NodeValue := TimeToStr(pi^.lastChange);
  end;
end;

procedure ExportBans(l: TList; FileName: String; asXML: boolean);
var
  xml: IXMLDocument;
  txt: TStringList;
  nbanlist: IXMLNode;
  i: Integer;
begin
  if asXML then
  begin
    xml := NewXMLDocument;
    xml.Encoding := 'UTF-8';
    xml.Options := xml.Options + [doNodeAutoIndent];
    xml.NodeIndentStr := ' ';
    nbanlist := xml.AddChild('banlist');
    CreateBanNodes(l, nbanlist);
    xml.SaveToFile(FileName);
  end else
  begin
    txt := TStringlist.Create;
    try
      for i := 0 to l.Count - 1 do
        txt.Add(Trim(PBan(l[i])^.nick + ' ' + PBan(l[i])^.comment));
      txt.SaveToFile(FileName);
    finally
      txt.Free;
    end;
  end;
end;

procedure ExportInfos(l: TList; FileName: String; asXML: boolean);
var
  xml: IXMLDocument;
  txt: TStringList;
  ninfolist: IXMLNode;
  i: Integer;
begin
  if asXML then
  begin
    xml := NewXMLDocument;
    xml.Encoding := 'UTF-8';
    xml.Options := xml.Options + [doNodeAutoIndent];
    xml.NodeIndentStr := ' ';
    ninfolist := xml.AddChild('infolist');
    createInfoNodes(l, ninfolist);
    xml.SaveToFile(FileName);
  end else
  begin
    txt := TStringlist.Create;
    try
      for i := 0 to l.Count - 1 do
        txt.Add(Trim(PBan(l[i])^.nick + ' ' + PBan(l[i])^.comment));
      txt.SaveToFile(FileName);
    finally
      txt.Free;
    end;
  end;
end;


end.
