unit UMessageCenter;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, Buttons, inifile, XMLDoc, XMLIntf, WC3Types,
  XMLDom, RegExpr;

type

  TMessageChange = procedure(kind: TWhisperType) of object;

  TMessageCenter = class(TObject)
  private
    XML: IXMLDocument;
    FMainNode: IXMLNode;
    FFileName: TFileName;
    FCurrentNode: IXMLNode;
    FChange: TMessageChange;
    FStyle: string;
    FFirstNotePresent: Boolean;
    procedure MarkLinks(node: IXMLNode);
  public
    constructor Create(Parent: TControl; FileName: TFileName; Style: string);
    procedure SaveToFile;
    procedure AddChat(nick: string; text: string; kind: TWhisperType);
    procedure AddInfo(text: string; kind: TWhisperType);
    procedure AddError(text, job: string; debug: string);
    procedure ChangeChannel(channel: string);
    procedure ChangeGame(game: string; map: string);
    property OnChange: TMessageChange read FChange write FChange;
    procedure SetStyleSheet(sheet: string);
    procedure AddFirstNote;
    procedure RemoveFirstNote;
    procedure Clear;
    procedure CutToSize(messagesCount: Integer);
    procedure Init;
  end;

var
  MessageCenter: TMessageCenter;

const
  SWelcomeNote =
    'Here you will find all messages whispered. No messages received yet.';

implementation

{ MessageCenter }

procedure TMessageCenter.AddChat(nick: string; text: string; kind:
  TWhisperType);
var
  msg: IXMLNode;
begin
  RemoveFirstNote;
  msg := FCurrentNode.AddChild('message');
  case kind of
    wtIn: msg.Attributes['type'] := 'whisperin';
    wtOut: msg.Attributes['type'] := 'whisperout';
    wtEmoteIn: msg.Attributes['type'] := 'emotein';
    wtEmoteOut: msg.Attributes['type'] := 'emoteout';
    wtChatIn: msg.Attributes['type'] := 'chatin';
    wtChatOut: msg.Attributes['type'] := 'chatout';
    wtChatBlizz: msg.Attributes['type'] := 'chatblizz';
  else
    raise Exception.Create('Unknown chat type supplied.' + cr + nick + cr +
      text);
  end;
  msg.AddChild('nick').Text := Utf8ToAnsi(nick);
  msg.AddChild('text').Text := Utf8ToAnsi(text);
  msg.AddChild('time').Text := DateTimeToStr(Now);
  MarkLinks(msg.ChildNodes['text']);
  CutToSize(100);
  OnChange(kind);
end;

procedure TMessageCenter.AddInfo(text: string; kind: TWhisperType);
var
  msg: IXMLNode;
begin
  RemoveFirstNote;
  msg := FCurrentNode.AddChild('message');
  case kind of
    wtBNetInfo: msg.Attributes['type'] := 'bnetinfo';
    wtBNetError: msg.Attributes['type'] := 'bneterror';
    wtNote: msg.Attributes['type'] := 'note';
  else
    raise Exception.Create('Unknown info type supplied.' + cr + text);
  end;
  msg.AddChild('text').Text := Utf8ToAnsi(text);
  msg.AddChild('time').Text := DateTimeToStr(Now);
  MarkLinks(msg.ChildNodes['text']);
  if Assigned(OnChange) then
    OnChange(kind);
end;

procedure TMessageCenter.ChangeChannel(channel: string);
begin
  RemoveFirstNote;
  FCurrentNode := FMainNode.AddChild('channel');
  FCurrentNode.Attributes['title'] := channel;
  OnChange(wtNull);
end;

procedure TMessageCenter.ChangeGame(game, map: string);
begin
  RemoveFirstNote;
  FCurrentNode := FMainNode.AddChild('game');
  FCurrentNode.Attributes['title'] := game;
  FCurrentNode.Attributes['mapname'] := map;
  OnChange(wtNull);
end;

constructor TMessageCenter.Create(Parent: TControl; FileName: TFileName; Style:
  string);
begin
  FFileName := FileName;
  FStyle := Style;
  try
    XML := LoadXMLDocument(FFileName);
    FMainNode := XML.DocumentElement;
    FCurrentNode := FMainNode;
  except
    DeleteFile(FFileName);
    XML := NewXMLDocument;
    Init;
    AddFirstNote;
  end;
end;

procedure TMessageCenter.SaveToFile;
var
  i: Integer;
  n: IXMLNode;
begin
  with XML do
    for i := ChildNodes.Count - 1 downto 0 do
    begin
      n := ChildNodes.Nodes[i];
      if (n.NodeType = ntProcessingInstr) and (n.NodeName = 'xml-stylesheet')
        then
        n.ParentNode.ChildNodes.Remove(n);
    end;
  n := XML.CreateNode('xml-stylesheet', ntProcessingInstr,
    'type="text/xsl" href="' + FStyle + '"');
  XML.ChildNodes.Insert(1, n);
  XML.SaveToFile(FFileName);
end;

procedure TMessageCenter.SetStyleSheet(sheet: string);
begin
  FStyle := sheet;
  OnChange(wtStyle);
end;

procedure TMessageCenter.Clear;
begin
  XML.ChildNodes.Clear;
  XML.Resync;
  Init;
end;

procedure TMessageCenter.Init;
begin
  XML.Encoding := 'UTF-8';
  XML.Options := XML.Options + [doNodeAutoIndent];
  FMainNode := XML.AddChild('chatlog');
  FCurrentNode := FMainNode;
  FFirstNotePresent := False;
end;

procedure TMessageCenter.AddError(text, job, debug: string);
var
  msg: IXMLNode;
  debugnode: IXMLNode;
begin
  RemoveFirstNote;
  msg := FCurrentNode.AddChild('message');
  msg.Attributes['type'] := 'error';
  msg.AddChild('text').Text := text;
  msg.AddChild('time').Text := DateTimeToStr(Now);
  debugnode := msg.AddChild('debug');
  debugnode.AddChild('data').Text := debug;
  debugnode.AddChild('job').Text := job;
  OnChange(wtError);
end;

procedure TMessageCenter.AddFirstNote;
begin
  AddInfo(SWelcomeNote, wtNote);
  FFirstNotePresent := True;
end;

procedure TMessageCenter.RemoveFirstNote;
var
  i: Integer;
begin
  if FFirstNotePresent then
  begin
    for i := 0 to FMainNode.ChildNodes.Count - 1 do
      if FMainNode.ChildNodes[i].Attributes['type'] = 'note' then
      begin
        FMainNode.ChildNodes.Delete(i);
        break;
      end;
    FFirstNotePresent := False;
  end;
end;

procedure TMessageCenter.MarkLinks(node: IXMLNode);
var
  a: string;
  temp: String;
  t: TRegExpr;
  n: IXMLNode;
  first: boolean;
const
  expr : String = '(.*)(^|\s)(http://.+?|ftp://.+?|www\..+?)(\s|\.+?\s|\!+?\s)(.*)';
begin
  a := ' '+node.Text+' ';
  node.Text := '';
  first := true;
  t := TRegExpr.Create;
  t.ModifierI := true;
  t.Expression := expr;
  while t.Exec(a) do
  begin
    temp := t.Match[4] + t.Match[5];
    if first then
      delete(temp, length(temp), 1);
    n := XML.CreateNode(temp, ntText);
    node.ChildNodes.Insert(0, n);
    n := XML.CreateNode('link');
    n.Text := t.Match[3];
    node.ChildNodes.Insert(0, n);
    a := t.Match[1] + t.Match[2];
    first := false;
  end;
  delete(a, 1, 1);
  if first then
    delete(a, length(a), 1);
  n := XML.CreateNode(a, ntText);
  node.ChildNodes.Insert(0, n);
  t.Free;
end;

procedure TMessageCenter.CutToSize(messagesCount: Integer);
var
  Count: Integer;
  n: IXMLNode;
  i: Integer;
begin
  with FMainNode do
  begin
    Count := 0;
    for i := 0 to ChildNodes.Count - 1 do
    begin
      if ChildNodes[i].NodeName = 'message' then inc(Count)
      else inc(Count, ChildNodes[i].ChildNodes.Count);
    end;
    dec(Count, MessagesCount);
    while Count > 0 do
    begin
      n := FMainNode.ChildNodes[0];
      if n.NodeName = 'message' then begin
        dec(count);
        FMainNode.ChildNodes.Delete(0);
      end else
      begin
        if n.ChildNodes.Count > 0 then
        begin
          n.ChildNodes.Delete(0);
          dec(Count);
        end else
          FMainNode.ChildNodes.Delete(0);
      end;
    end;
  end;
end;

end.
